#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import os
import re
import datetime
import functools
import json
import uuid

from .base import DatabaseBackend

from muninn._compat import dictkeys, dictvalues

# Select a version of dbapi2 that's available.
# Since we need the spatialite module, we need a version of sqlite3 that supports extensions.
# The defaulft version of sqlite3 that comes with Python does not support this, so we prefer pysqlite2
# (which does come with an extension-enabled version of sqlite3) over the sqlite3 python module.
try:
    import pysqlite2.dbapi2 as dbapi2
except ImportError:
    import sqlite3.dbapi2 as dbapi2

import muninn.config as config
import muninn.database.sql as sql
import muninn.database.blobgeometry as blobgeometry
import muninn.geometry as geometry

from muninn.exceptions import *
from muninn.function import Prototype
from muninn.schema import *
from muninn.struct import Struct


class _SQLiteConfig(Mapping):
    _alias = "sqlite"

    connection_string = Text
    mod_spatialite_path = optional(Text)
    table_prefix = optional(Text)


def create(configuration):
    options = config.parse(configuration.get("sqlite", {}), _SQLiteConfig)
    _SQLiteConfig.validate(options)
    return SQLiteBackend(**options)


class SQLiteError(Error):
    def __init__(self, message=None):
        message = "sqlite backend error" + ("" if not message else ": " + message)
        super(SQLiteError, self).__init__(message)


def translate_sqlite_errors(func):
    """Decorator that translates sqlite dbapi exceptions into muninn exceptions."""
    @functools.wraps(func)
    def translate_sqlite_errors_(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except dbapi2.Error as _error:
            try:
                message = _error.diag.message_primary
            except AttributeError:
                message = None

            if message:
                try:
                    message_detail = _error.diag.message_detail
                except AttributeError:
                    message_detail = None

                if message_detail:
                    message += " [" + message_detail + "]"
            else:
                # Remove newlines and excessive whitespace from the original SQLite exception message.
                message = " ".join(str(_error).split())

            raise SQLiteError(message)

    return translate_sqlite_errors_


def _adapt_geometry(geometry):
    """Return the SQLite BLOB-Geometry representation of a Geometry instance."""
    return dbapi2.Binary(blobgeometry.encode_blob_geometry(geometry))


def _cast_geometry(blob):
    """Construct a Geometry instance from its SQLite BLOB-Geometry representation."""
    if blob is None:
        return blob
    return blobgeometry.decode_blob_geometry(blob)


class SQLiteConnection(object):
    """Wrapper for a sqlite database connection that defers (re)connection until an attempt is made to use the
    connection.

    Only non-nested transactions are supported, no auto-commit or nested transactions. A transaction can be started
    using the context manager interface.

    """
    def __init__(self, connection_string, mod_spatialite_path, backend):
        self._connection_string = connection_string
        self._mod_spatialite = mod_spatialite_path
        self._connection = None
        self._in_transaction = False
        self._backend = backend

    def __enter__(self):
        # Begin a transaction. The transaction is not started immediately, but a state change is recorded such that
        # attempts to start nested transactions can be detected. Also, the connection with the database is
        # re-established if necessary.
        #
        if self._in_transaction:
            raise InternalError("nested transactions are not supported")

        # Reconnect if necessary.
        if self._connection is None:
            self._connect()

        # Change state to guard against nested transactions.
        self._in_transaction = True

        return self

    def __exit__(self, type, value, traceback):
        # End a transaction, either by committing the transaction, or by rolling back in case an exception occured.
        try:
            if type is None:
                self._connection.commit()
            else:
                self._connection.rollback()
        finally:
            self._in_transaction = False
            self.close()

    def _connect(self):
        # Re-establish the connection to the database.
        need_prepare = not os.path.exists(self._connection_string)
        self._connection = dbapi2.connect(self._connection_string, detect_types=dbapi2.PARSE_DECLTYPES)

        # make sure that foreign keys are enabled
        self._connection.execute("PRAGMA foreign_keys = ON;")

        # load the spatialite extension
        try:
            self._connection.enable_load_extension(True)
            self._connection.execute("select load_extension(\"%s\");" % (self._mod_spatialite))
        except Exception as e:
            raise Error("loading mod_spatialite extension failed (mod_spatialite_path='%s'): %s" %
                        (self._mod_spatialite, str(e)))

        # ensure that spatial metadata init has been done
        with self._connection:
            # enter auto-commit mode, to have complete control the transaction state
            # see https://docs.python.org/3/library/sqlite3.html#controlling-transactions
            isolation_level = self._connection.isolation_level
            self._connection.isolation_level = None
            cursor = self._connection.cursor()
            cursor.execute("SELECT CheckSpatialMetaData();")
            try:
                if cursor.fetchone()[0] == 0:
                    # execute InitSpatialMetadata inside a transaction, to make it faster
                    cursor.execute("BEGIN")
                    cursor.execute("SELECT InitSpatialMetadata()")
                    cursor.execute("COMMIT")
            finally:
                cursor.close()
                # restore isolation level
                self._connection.isolation_level = isolation_level

        # create the tables if necessary
        if need_prepare:
            with self._connection:
                # we need to perform a transaction before we allow the actual transaction to happen
                self._in_transaction = True
                sqls = self._backend._create_tables_sql()
                self._backend._execute_list(sqls)
                self._connection.commit()
                self._in_transaction = False

        # Make sure TEXT data is translated to UTF-8 str types in Python (to align with default psycopg2 behavior)
        self._connection.text_factory = str

    def _disconnect(self):
        self._connection.close()
        self._connection = None

    def close(self):
        if self._in_transaction:
            raise InternalError("unable to close the connection with the database while a transaction is in progress")

        # Close the connection.
        if self._connection is not None:
            self._disconnect()

    def cursor(self):
        if not self._in_transaction:
            raise InternalError("creating a cursor requires an active transaction")

        return self._connection.cursor()


class SQLiteBackend(object):
    def __init__(self, connection_string="", mod_spatialite_path="mod_spatialite", table_prefix=""):
        dbapi2.register_converter("BOOLEAN", lambda x: bool(int(x)))
        dbapi2.register_adapter(bool, lambda x: int(x))

        dbapi2.register_converter("UUID", lambda x: uuid.UUID(x.decode()))
        dbapi2.register_adapter(uuid.UUID, lambda x: x.hex)

        dbapi2.register_converter("GEOMETRY", _cast_geometry)
        dbapi2.register_adapter(geometry.Point, _adapt_geometry)
        dbapi2.register_adapter(geometry.LineString, _adapt_geometry)
        dbapi2.register_adapter(geometry.Polygon, _adapt_geometry)
        dbapi2.register_adapter(geometry.MultiPoint, _adapt_geometry)
        dbapi2.register_adapter(geometry.MultiLineString, _adapt_geometry)
        dbapi2.register_adapter(geometry.MultiPolygon, _adapt_geometry)

        self._connection_string = connection_string
        self._connection = SQLiteConnection(connection_string, mod_spatialite_path, self)

        if table_prefix and not re.match(r"[a-z][_a-z]*(\.[a-z][_a-z]*)*", table_prefix):
            raise ValueError("invalid table_prefix %s" % table_prefix)
        self._table_prefix = table_prefix

        self._core_table_name = self._table_name("core")
        self._link_table_name = self._table_name("link")
        self._tag_table_name = self._table_name("tag")

        self._namespace_schemas = {}
        self._sql_builder = sql.SQLBuilder({}, sql.TypeMap(), {}, self._table_name, self._placeholder,
                                           self._placeholder, self._rewriter_property)

    def _create_tables_sql(self):
        result = []
        # Create the table for the core namespace.
        column_sql = []
        schema = self._namespace_schema("core")
        for name in schema:
            type_name = self._type_map()[schema[name]]
            if type_name != "GEOMETRY":
                sql = name + " " + type_name
                if not schema.is_optional(name):
                    sql = sql + " " + "NOT NULL"
                column_sql.append(sql)
        column_sql.append("PRIMARY KEY (uuid)")
        column_sql.append("UNIQUE (archive_path, physical_name)")
        column_sql.append("UNIQUE (product_type, product_name)")
        result.append("CREATE TABLE %s (%s);" % (self._core_table_name, ", ".join(column_sql)))
        for name in schema:
            if self._type_map()[schema[name]] == "GEOMETRY":
                result.append("SELECT AddGeometryColumn('%s', '%s', 4326, 'GEOMETRY', 2);" %
                              (self._core_table_name, name,))
        for name in schema:
            if schema.has_index(name):
                if schema[name] == Geometry:
                    # For the geospatial footprint we need to use a special spatial index
                    result.append("SELECT CreateSpatialIndex('%s', '%s');" % (self._core_table_name, name))
                else:
                    result.append("CREATE INDEX idx_%s_%s ON %s (%s);" %
                                  (self._core_table_name, name, self._core_table_name, name))

        # Create the tables for all non-core namespaces.
        for namespace in self._namespace_schemas:
            if namespace == "core":
                continue

            column_sql = []
            schema = self._namespace_schema(namespace)
            for name in schema:
                type_name = self._type_map()[schema[name]]
                if type_name != "GEOMETRY":
                    sql = name + " " + type_name
                    if not schema.is_optional(name):
                        sql = sql + " " + "NOT NULL"
                    column_sql.append(sql)
            column_sql.append("uuid UUID PRIMARY KEY REFERENCES %s(uuid) ON DELETE CASCADE" %
                              self._core_table_name)
            result.append("CREATE TABLE %s (%s);" % (self._table_name(namespace), ", ".join(column_sql)))
            for name in schema:
                if self._type_map()[schema[name]] == "GEOMETRY":
                    result.append("SELECT AddGeometryColumn('%s', '%s', 4326, 'GEOMETRY', 2);" %
                                  (self._table_name(namespace), name))
            for name in schema:
                if schema.has_index(name):
                    if schema[name] == Geometry:
                        result.append("SELECT CreateSpatialIndex('%s', '%s');" % (self._table_name(namespace), name))
                    else:
                        result.append("CREATE INDEX idx_%s_%s ON %s (%s);" %
                                      (self._table_name(namespace), name, self._table_name(namespace), name))

        # We use explicit 'id' primary keys for the links and tags tables so the entries can be managed using
        # other front-ends that may not support tuples as primary keys.

        # Create the table for links.
        result.append("CREATE TABLE %s (id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, "
                      "uuid UUID REFERENCES %s(uuid) ON DELETE CASCADE, "
                      "source_uuid UUID NOT NULL, UNIQUE (uuid, source_uuid));" %
                      (self._link_table_name, self._core_table_name))
        result.append("CREATE INDEX idx_%s_uuid ON %s (uuid);" % (self._link_table_name, self._link_table_name))
        result.append("CREATE INDEX idx_%s_source_uuid ON %s (source_uuid);" %
                      (self._link_table_name, self._link_table_name))

        # Create the table for tags.
        result.append("CREATE TABLE %s (id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, "
                      "uuid UUID REFERENCES %s(uuid) ON DELETE CASCADE, "
                      "tag TEXT NOT NULL, UNIQUE (uuid, tag));" % (self._tag_table_name, self._core_table_name))
        result.append("CREATE INDEX idx_%s_uuid ON %s (uuid);" % (self._tag_table_name, self._tag_table_name))
        result.append("CREATE INDEX idx_%s_tag ON %s (tag);" % (self._tag_table_name, self._tag_table_name))
        return result

    def _delete_product_properties(self, uuid):
        cursor = self._connection.cursor()
        try:
            cursor.execute("DELETE FROM %s WHERE source_uuid = %s" % (self._link_table_name, self._placeholder()),
                           (uuid,))
            cursor.execute("DELETE FROM %s WHERE uuid = %s" % (self._core_table_name, self._placeholder()), (uuid,))
            assert cursor.rowcount <= 1

            if cursor.rowcount != 1:
                raise Error("could not delete properties for product: %s" % (uuid,))
        finally:
            cursor.close()

    def _derived_products(self, uuid):
        query = "SELECT uuid FROM %s WHERE source_uuid = %s" % (self._link_table_name, self._placeholder())
        parameters = (uuid,)

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
            return [row[0] for row in cursor]
        finally:
            cursor.close()

    def _drop_tables(self):
        with self._connection:
            cursor = self._connection.cursor()
            try:
                # remove geospatial footprint index
                cursor.execute("SELECT DisableSpatialIndex('%s', '%s')" % (self._core_table_name, 'footprint'))
                cursor.execute("DROP TABLE IF EXISTS idx_%s_footprint" % self._core_table_name)
                # first remote all links to geometry columns
                for namespace in self._namespace_schemas:
                    schema = self._namespace_schema(namespace)
                    for name in schema:
                        if self._type_map()[schema[name]] == "GEOMETRY":
                            cursor.execute("SELECT DiscardGeometryColumn('%s', '%s')" %
                                           (self._table_name(namespace), name))
                # then remove the tables
                cursor.execute("DROP TABLE IF EXISTS %s" % self._tag_table_name)
                cursor.execute("DROP TABLE IF EXISTS %s" % self._link_table_name)
                for namespace in self._namespace_schemas:
                    if namespace != "core":
                        cursor.execute("DROP TABLE IF EXISTS %s" % self._table_name(namespace))
                # remove 'core' table last
                if "core" in self._namespace_schemas:
                    cursor.execute("DROP TABLE IF EXISTS %s" % self._core_table_name)
            finally:
                cursor.close()

    def _execute_list(self, sql_list):
        cursor = self._connection.cursor()
        try:
            for item in sql_list:
                cursor.execute(item)
        finally:
            cursor.close()

    def _find_products_without_available_source(self, product_type=None, grace_period=datetime.timedelta()):
        core_properties = list(self._namespace_schema("core"))
        select_list = ["%s.%s" % (self._core_table_name, name) for name in core_properties]

        query = "SELECT %s FROM %s WHERE active AND strftime('%%s', 'now') - strftime('%%s', archive_date) > %s AND " \
                "uuid IN (SELECT uuid FROM %s EXCEPT SELECT DISTINCT link.uuid FROM %s AS link LEFT JOIN %s AS " \
                "source ON (link.source_uuid = source.uuid) WHERE source.uuid IS NULL OR source.archive_path IS NOT " \
                "NULL)" % (", ".join(select_list), self._core_table_name, self._placeholder(), self._link_table_name,
                           self._link_table_name, self._core_table_name)

        if product_type is not None:
            query = "%s AND product_type = %s" % (query, self._placeholder())

        grace_period = grace_period.total_seconds()

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, (grace_period,) if product_type is None else (grace_period, product_type))

            description = [("core", core_properties)]
            return [self._unpack_product_properties(description, row) for row in cursor]
        finally:
            cursor.close()

    def _find_products_without_source(self, product_type=None, grace_period=datetime.timedelta(),
                                      archived_only=False):
        core_properties = list(self._namespace_schema("core"))
        select_list = ["%s.%s" % (self._core_table_name, name) for name in core_properties]
        query = "SELECT %s FROM %s WHERE %s.active AND strftime('%%s', 'now') - strftime('%%s', %s.archive_date) > " \
                "%s AND NOT EXISTS (SELECT 1 FROM %s WHERE %s.uuid = %s.uuid)" % \
                (", ".join(select_list), self._core_table_name, self._core_table_name, self._core_table_name,
                 self._placeholder(), self._link_table_name, self._link_table_name, self._core_table_name)

        if product_type is not None:
            query = "%s AND product_type = %s" % (query, self._placeholder())

        if archived_only:
            query = "%s AND archive_path IS NOT NULL" % query

        grace_period = grace_period.total_seconds()

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, (grace_period,) if product_type is None else (grace_period, product_type))

            description = [("core", core_properties)]
            return [self._unpack_product_properties(description, row) for row in cursor]
        finally:
            cursor.close()

    def _insert_namespace_properties(self, uuid, name, properties):
        self._validate_namespace_properties(name, properties)
        assert uuid is not None and getattr(properties, "uuid", uuid) == uuid

        # Split the properties into a list of (database) field names and a list of values. This assumes the database
        # field that corresponds to a given property has the same name. If the backend uses different field names, the
        # required translation can be performed here. Values can also be translated if necessary.
        properties_dict = vars(properties)
        fields, parameters = dictkeys(properties_dict), dictvalues(properties_dict)
        # Ensure the uuid field is present (for namespaces other than the core namespace this is used as the foreign
        # key).
        if "uuid" not in properties:
            fields.append("uuid")
            parameters.append(uuid)

        parameters = [json.dumps(p) if isinstance(p, dict) else p for p in parameters]

        # Build and execute INSERT query.
        query = "INSERT INTO %s (%s) VALUES (%s)" % (self._table_name(name), ", ".join(fields),
                                                     ", ".join([self._placeholder()] * len(fields)))

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
        finally:
            cursor.close()

    def _delete_namespace_properties(self, uuid, name):
        query = "DELETE FROM %s WHERE uuid=%s" % (self._table_name(name), self._placeholder())
        cursor = self._connection.cursor()
        try:
            cursor.execute(query, [uuid])
        finally:
            cursor.close()

    def _link(self, uuid, source_uuids):
        query = "INSERT OR IGNORE INTO %s (uuid, source_uuid) VALUES (%s, %s)" % \
            (self._link_table_name, self._placeholder(), self._placeholder())

        for source_uuid in source_uuids:
            with self._connection:
                cursor = self._connection.cursor()
                try:
                    cursor.execute(query, (uuid, source_uuid))
                finally:
                    cursor.close()

    def _namespace_schema(self, namespace):
        try:
            return self._namespace_schemas[namespace]
        except KeyError:
            raise Error("undefined namespace: \"%s\"" % namespace)

    def _placeholder(self, name=None, arg=None):
        if name is not None:
            return ":%s" % name
        else:
            return "?"

    def _rewriter_property(self, column_name, subscript):
        # timestamp
        if subscript == 'year':
            return "STRFTIME('%%Y', %s)" % column_name
        if subscript == 'month':
            return "STRFTIME('%%m', %s)" % column_name
        if subscript == 'yearmonth':
            return "STRFTIME('%%Y-%%m', %s)" % column_name
        if subscript == 'day':
            return "STRFTIME('%%d', %s)" % column_name
        if subscript == 'date':
            return "STRFTIME('%%Y-%%m-%%d', %s)" % column_name
        if subscript == 'hour':
            return "STRFTIME('%%H', %s)" % column_name
        if subscript == 'minute':
            return "STRFTIME('%%M', %s)" % column_name
        if subscript == 'second':
            return "STRFTIME('%%S', %s)" % column_name
        if subscript == 'time':
            return "STRFTIME('%%H:%%M:%%S', %s)" % column_name
        # text
        if subscript == 'length':
            return "LENGTH(%s)" % column_name
        raise ValueError('Unsupported subscript: %s' % subscript)

    def _rewriter_table(self):
        rewriter_table = sql.default_rewriter_table()

        #
        # Timestamp binary minus operator.
        #
        rewriter_table[Prototype("-", (Timestamp, Timestamp), Real)] = \
            lambda arg0, arg1: "(julianday(%s) - julianday(%s)) * 86400.0" % (arg0, arg1)

        #
        # Enable escape sequences with the LIKE operator
        #
        rewriter_table[Prototype("~=", (Text, Text), Boolean)] = \
            lambda arg0, arg1: "(%s) LIKE (%s) ESCAPE '\\'" % (arg0, arg1)

        #
        # Functions.
        #
        rewriter_table[Prototype("covers", (Geometry, Geometry), Boolean)] = \
            lambda arg0, arg1: "(ST_Covers(%s, %s)=1)" % (arg0, arg1)

        rewriter_table[Prototype("distance", (Geometry, Geometry), Real)] = \
            lambda arg0, arg1: "ST_Distance(%s, %s)" % (arg0, arg1)

        rewriter_table[Prototype("intersects", (Geometry, Geometry), Boolean)] = \
            lambda arg0, arg1: "(ST_Intersects(%s, %s)=1)" % (arg0, arg1)

        rewriter_table[Prototype("is_source_of", (UUID,), Boolean)] = \
            lambda arg0: "EXISTS (SELECT 1 FROM %s WHERE source_uuid = %s.uuid AND uuid = (%s))" % \
            (self._link_table_name, self._core_table_name, arg0)

        def is_source_of_subquery(where_expr, where_namespaces):
            joins = ''
            for namespace in where_namespaces:
                joins = "%s INNER JOIN %s USING (uuid)" % (joins, self._table_name(namespace))

            return "{core}.uuid in (SELECT {link}.source_uuid FROM {core} {joins} " \
                "INNER JOIN {link} on {link}.uuid = {core}.uuid WHERE {where})".format(
                    core=self._core_table_name, link=self._link_table_name, joins=joins,
                    where=where_expr)

        rewriter_table[Prototype("is_source_of", (Boolean,), Boolean)] = is_source_of_subquery

        rewriter_table[Prototype("is_derived_from", (UUID,), Boolean)] = \
            lambda arg0: "EXISTS (SELECT 1 FROM %s WHERE uuid = %s.uuid AND source_uuid = (%s))" % \
            (self._link_table_name, self._core_table_name, arg0)

        def is_derived_from_subquery(where_expr, where_namespaces):
            joins = ''
            for namespace in where_namespaces:
                joins = "%s INNER JOIN %s USING (uuid)" % (joins, self._table_name(namespace))

            return "{core}.uuid in (SELECT {link}.uuid FROM {core} {joins} " \
                "INNER JOIN {link} on {link}.source_uuid = {core}.uuid WHERE {where})".format(
                    core=self._core_table_name, link=self._link_table_name, joins=joins,
                    where=where_expr)

        rewriter_table[Prototype("is_derived_from", (Boolean,), Boolean)] = is_derived_from_subquery

        rewriter_table[Prototype("has_tag", (Text,), Boolean)] = \
            lambda arg0: "EXISTS (SELECT 1 FROM %s WHERE uuid = %s.uuid AND tag = (%s))" % \
            (self._tag_table_name, self._core_table_name, arg0)

        rewriter_table[Prototype("now", (), Timestamp)] = \
            sql.as_is("datetime(\"now\")")

        def is_defined_rewriter(arg):
            namespace_name = arg.split('.')
            if len(namespace_name) == 1:
                return 'EXISTS (SELECT 1 FROM %s WHERE uuid = %s.uuid)' % \
                    (self._table_name(arg), self._core_table_name)
            else:
                return "(%s) IS NOT NULL" % arg

        rewriter_table[Prototype("is_defined", (Long,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Integer,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Real,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Boolean,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Text,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Namespace,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Timestamp,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (UUID,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Geometry,), Boolean)] = is_defined_rewriter

        return rewriter_table

    def _source_products(self, uuid):
        query = "SELECT source_uuid FROM %s WHERE uuid = %s" % (self._link_table_name, self._placeholder())
        parameters = (uuid,)

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
            return [row[0] for row in cursor]
        finally:
            cursor.close()

    def _table_name(self, name):
        return name if not self._table_prefix else self._table_prefix + name

    def _tag(self, uuid, tags):
        query = "INSERT OR IGNORE INTO %s (uuid, tag) VALUES (%s, %s)" % \
            (self._tag_table_name, self._placeholder(), self._placeholder())
        for tag in tags:
            with self._connection:
                cursor = self._connection.cursor()
                try:
                    cursor.execute(query, (uuid, tag))
                finally:
                    cursor.close()

    def _tags(self, uuid):
        query = "SELECT tag FROM %s WHERE uuid = %s ORDER BY tag" % (self._tag_table_name, self._placeholder())
        parameters = (uuid,)

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
            return [row[0] for row in cursor]
        finally:
            cursor.close()

    def _type_map(self):
        type_map = sql.TypeMap()
        type_map[Long] = "INTEGER"
        type_map[Integer] = "INTEGER"
        type_map[Real] = "REAL"
        type_map[Boolean] = "BOOLEAN"
        type_map[Text] = "TEXT"
        type_map[Timestamp] = "TIMESTAMP"
        type_map[UUID] = "UUID"
        type_map[Geometry] = "GEOMETRY"
        type_map[JSON] = "TEXT"

        return type_map

    def _unlink(self, uuid, source_uuids=None):
        if source_uuids is None:
            query = "DELETE FROM %s WHERE uuid = %s" % (self._link_table_name, self._placeholder())
            parameters = (uuid,)
        else:
            query = "DELETE FROM %s WHERE uuid = %s AND source_uuid IN (%s)" % \
                (self._link_table_name, self._placeholder(), ','.join(self._placeholder() * len(source_uuids)))
            parameters = [uuid] + list(source_uuids)

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
        finally:
            cursor.close()

    def _unpack_product_properties(self, description, values):
        unpacked_properties, start = Struct(), 0
        for ns_name, ns_description in description:
            end = start + len(ns_description)

            # A value of None for the uuid field of namespaces other than the core namespace indicates the namespace is
            # not defined for the product under consideration. In this case, the entire namespace will be skipped.
            # Otherwise, only the uuid field is skipped, as it is an implementation detail (foreign key) and it is not
            # part of the namespace itself.
            #
            if ns_name != "core":
                assert ns_description[0] == "uuid"
                if values[start] is None:
                    # Skip the entire namespace.
                    start = end
                    continue

                # Skip the uuid field.
                start += 1
                ns_description = ns_description[1:]

            unpacked_ns_properties = self._unpack_namespace_properties(ns_name, ns_description, values[start:end])
            self._validate_namespace_properties(ns_name, unpacked_ns_properties, partial=True)
            unpacked_properties[ns_name] = unpacked_ns_properties
            start = end

        return unpacked_properties

    def _unpack_namespace_properties(self, namespace, description, values):
        unpacked_properties = Struct()
        schema = self._namespace_schema(namespace)
        for identifier, value in zip(description, values):
            if value is not None or not schema.is_optional(identifier):
                if issubclass(schema[identifier], JSON):
                    value = json.loads(value)
                unpacked_properties[identifier] = value
        return unpacked_properties

    def _untag(self, uuid, tags=None):
        if tags is None:
            query = "DELETE FROM %s WHERE uuid = %s" % (self._tag_table_name, self._placeholder())
            parameters = (uuid,)
        else:
            query = "DELETE FROM %s WHERE uuid = %s AND tag IN (%s)" % \
                    (self._tag_table_name, self._placeholder(), ','.join(self._placeholder() * len(tags)))
            parameters = [uuid] + list(tags)

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
        finally:
            cursor.close()

    def _update_namespace_properties(self, uuid, name, properties):
        self._validate_namespace_properties(name, properties, partial=True)
        assert uuid is not None and getattr(properties, "uuid", uuid) == uuid

        # Split the properties into a list of (database) field names and a list of values. This assumes the database
        # field that corresponds to a given property has the same name. If the backend uses different field names, the
        # required translation can be performed here. Values can also be translated if necessary.
        properties_dict = vars(properties)
        fields, parameters = dictkeys(properties_dict), dictvalues(properties_dict)
        if not fields:
            return  # nothing to do

        # Remove the uuid field if present. This field needs to be included in the WHERE clause of the UPDATE query,
        # not in the SET clause.
        try:
            uuid_index = fields.index("uuid")
        except ValueError:
            pass
        else:
            del fields[uuid_index]
            del parameters[uuid_index]

        # Append the uuid (value) at the end of the list of parameters (will be used in the WHERE clause).
        parameters.append(uuid)

        parameters = [json.dumps(p) if isinstance(p, dict) else p for p in parameters]

        # Build and execute UPDATE query.
        set_clause = ", ".join(["%s = %s" % (field, self._placeholder()) for field in fields])
        query = "UPDATE %s SET %s WHERE uuid = %s" % (self._table_name(name), set_clause, self._placeholder())

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
            assert cursor.rowcount <= 1

            if cursor.rowcount != 1:
                raise Error("could not update properties for namespace: %s for product: %s" % (name, uuid))
        finally:
            cursor.close()

    def _validate_namespace_properties(self, namespace, properties, partial=False):
        self._namespace_schema(namespace).validate(properties, partial)

    @translate_sqlite_errors
    def count(self, where="", parameters={}):
        query, query_parameters = self._sql_builder.build_count_query(where, parameters)

        with self._connection:
            cursor = self._connection.cursor()
            try:
                cursor.execute(query, query_parameters)
                return cursor.fetchone()[0]
            finally:
                cursor.close()

    @translate_sqlite_errors
    def delete_product_properties(self, uuid):
        with self._connection:
            self._delete_product_properties(uuid)

    @translate_sqlite_errors
    def derived_products(self, uuid):
        with self._connection:
            return self._derived_products(uuid)

    @translate_sqlite_errors
    def destroy(self):
        # Each table is deleted in a separate transaction, such that the exception that is raised when a table does not
        # exist can be ignored (this requires the transaction to be rolled back).
        self._drop_tables()

    @translate_sqlite_errors
    def disconnect(self):
        """Drop the connection to the database in order to free up resources. The connection will be re-established
        automatically when required."""
        self._connection.close()

    def exists(self):
        if not os.path.isfile(self._connection_string):
            return False
        with self._connection:
            query = "SELECT name FROM sqlite_master WHERE type='table' AND name=%s" % (self._placeholder(),)
            parameters = (self._core_table_name,)
            cursor = self._connection.cursor()
            try:
                cursor.execute(query, parameters)
                return len(cursor.fetchall()) != 0
            finally:
                cursor.close()

    @translate_sqlite_errors
    def find_products_without_available_source(self, product_type=None, grace_period=datetime.timedelta()):
        """Return the core properties of all products that are linked to one or more source products, all of which are
           unavailable. A product is unavailable if there is no data associated with it, only properties. Products that
           have links to external source products will not be selected by this function, because it cannot be
           determined whether or not these products are available.

           Keyword arguments:
           product_type --  Only consider products of the specified product type.

        """
        with self._connection:
            return self._find_products_without_available_source(product_type, grace_period)

    @translate_sqlite_errors
    def find_products_without_source(self, product_type=None, grace_period=datetime.timedelta(),
                                     archived_only=False):
        """Return the core properties of all products that are not linked to any source products.

           Keyword arguments:
           product_type --  Only consider products of the specified product type.

        """
        with self._connection:
            return self._find_products_without_source(product_type, grace_period, archived_only)

    def initialize(self, namespace_schemas):
        self._namespace_schemas = namespace_schemas
        self._sql_builder = sql.SQLBuilder(self._namespace_schemas, self._type_map(), self._rewriter_table(),
                                           self._table_name, self._placeholder, self._placeholder,
                                           self._rewriter_property)

    @translate_sqlite_errors
    def insert_product_properties(self, properties):
        with self._connection:
            self._insert_namespace_properties(properties.core.uuid, "core", properties.core)
            for ns_name, ns_properties in vars(properties).items():
                if ns_name == "core" or ns_properties is None:
                    continue
                self._insert_namespace_properties(properties.core.uuid, ns_name, ns_properties)

    @translate_sqlite_errors
    def link(self, uuid, source_uuids):
        self._link(uuid, source_uuids)

    @translate_sqlite_errors
    def prepare(self, dry_run=False):
        sqls = self._create_tables_sql()
        if not dry_run:
            if os.path.isfile(self._connection_string):
                with self._connection:
                    self._execute_list(sqls)
            else:
                # If the db did not exist before, then trigger the table creation by simply making a connection
                with self._connection:
                    pass
        return sqls

    @translate_sqlite_errors
    def search(self, where="", order_by=[], limit=None, parameters={}, namespaces=[], property_names=[]):
        query, query_parameters, query_description = \
            self._sql_builder.build_search_query(where, order_by, limit, parameters, namespaces, property_names)

        with self._connection:
            cursor = self._connection.cursor()
            try:
                cursor.execute(query, query_parameters)
                return [self._unpack_product_properties(query_description, row) for row in cursor]
            finally:
                cursor.close()

    @translate_sqlite_errors
    def server_time_utc(self):
        return datetime.datetime.utcnow()

    @translate_sqlite_errors
    def source_products(self, uuid):
        with self._connection:
            return self._source_products(uuid)

    @translate_sqlite_errors
    def summary(self, where="", parameters=None, aggregates=None, group_by=None, group_by_tag=False,
                having=None, order_by=None):
        query, query_parameters, query_description = self._sql_builder.build_summary_query(
            where, parameters, aggregates, group_by, group_by_tag, having, order_by)

        with self._connection:
            cursor = self._connection.cursor()
            try:
                cursor.execute(query, query_parameters)
                return [row for row in cursor], query_description
            finally:
                cursor.close()

    @translate_sqlite_errors
    def tag(self, uuid, tags):
        self._tag(uuid, tags)

    @translate_sqlite_errors
    def tags(self, uuid):
        with self._connection:
            return self._tags(uuid)

    @translate_sqlite_errors
    def update_product_properties(self, properties, uuid=None, new_namespaces=None):
        new_namespaces = new_namespaces or []
        if "core" in properties:
            self._validate_namespace_properties("core", properties.core, partial=True)
            if "uuid" in properties.core:
                uuid = properties.core.uuid if uuid is None else uuid
                if uuid != properties.core.uuid:
                    raise Error("specified uuid does not match uuid included in the specified product properties")

        if uuid is None:
            raise Error("no uuid specified and no uuid included in the specified product properties")

        with self._connection:
            self._update_namespace_properties(uuid, "core", properties.core)
            for ns_name, ns_properties in vars(properties).items():
                if ns_name == "core":
                    continue
                if ns_name in new_namespaces:
                    if ns_properties is not None:
                        self._insert_namespace_properties(uuid, ns_name, ns_properties)
                elif ns_properties is None:
                    self._delete_namespace_properties(uuid, ns_name)
                else:
                    self._update_namespace_properties(uuid, ns_name, ns_properties)

    @translate_sqlite_errors
    def unlink(self, uuid, source_uuids=None):
        with self._connection:
            self._unlink(uuid, source_uuids)

    @translate_sqlite_errors
    def untag(self, uuid, tags=None):
        with self._connection:
            self._untag(uuid, tags)



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function
from muninn._compat import string_types as basestring

import collections
import copy
import datetime
import functools
import inspect
import hashlib
import os
import re
import sys
from uuid import UUID, uuid4
import warnings

import muninn.config as config
import muninn.util as util

from muninn.core import Core
from muninn.exceptions import Error, StorageError
from muninn.extension import CascadeRule
from muninn.schema import Text, Integer, Sequence, Mapping
from muninn.struct import Struct
from muninn import remote

HASH_ALGORITHMS = set(hashlib.algorithms_guaranteed)


class _ExtensionName(Text):
    _alias = "extension_name"

    @classmethod
    def validate(cls, value):
        super(_ExtensionName, cls).validate(value)
        if not re.match(r"[a-zA-Z][_a-zA-Z0-9]*(\.[a-zA-Z][_a-zA-Z0-9]*)*$", value):
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class _ExtensionList(Sequence):
    _alias = "extension_list"
    sub_type = _ExtensionName


class _ArchiveConfig(Mapping):
    _alias = "archive"

    database = Text()
    storage = Text()
    cascade_grace_period = Integer(optional=True)
    max_cascade_cycles = Integer(optional=True)
    namespace_extensions = _ExtensionList(optional=True)
    product_type_extensions = _ExtensionList(optional=True)
    remote_backend_extensions = _ExtensionList(optional=True)
    hook_extensions = _ExtensionList(optional=True)
    auth_file = Text(optional=True)
    tempdir = Text(optional=True)


def _load_database_module(name):
    module_name = "muninn.database.%s" % name

    try:
        __import__(module_name)
    except ImportError as e:
        raise Error("import of database %r (module %r) failed (%s)" % (name, module_name, e))

    return sys.modules[module_name]


def _load_storage_module(name):
    module_name = "muninn.storage.%s" % name

    try:
        __import__(module_name)
    except ImportError as e:
        raise Error("import of storage %r (module %r) failed (%s)" % (name, module_name, e))

    return sys.modules[module_name]


def _load_extension(name):
    try:
        __import__(name)
    except ImportError as e:
        raise Error("import of extension %r failed (%s)" % (name, e))

    return sys.modules[name]


def _inspect_nargs(func):
    try:
        getargspec = inspect.getfullargspec
    except AttributeError:
        getargspec = inspect.getargspec

    return len(getargspec(func).args)


def _register_extension(archive, extension, name, plugin_type, list_name=None, get_name=None):
    if list_name is None:
        list_name = plugin_type + 's'
    if get_name is None:
        get_name = plugin_type

    try:
        list_function = getattr(extension, list_name)
        get_function = getattr(extension, get_name)
    except AttributeError:
        raise Error("extension %r does not implement the %s extension API" % (name, plugin_type))

    config_section = archive._configuration.get('extension:' + extension.__name__)

    list_args = []
    if _inspect_nargs(list_function) == 1:
        list_args.append(config_section)

    for plugin_name in list_function(*list_args):

        get_args = [plugin_name]
        if _inspect_nargs(get_function) == 2:
            get_args.append(config_section)

        plugin = get_function(*get_args)
        getattr(archive, 'register_' + plugin_type)(plugin_name, plugin)


_CORE_PROP_NAMES = [
    'uuid',
    'active',
    'product_name',
    'archive_path',
    'physical_name',
    'product_type',
]


class Archive(object):
    """Archive class

    The Archive class is used to represent and interact with Muninn archives.
    It provides functionality such as querying existing or ingesting new
    products. While at the core of the Muninn command-line tools, it can also
    be used directly.

    It is typically instantiated and used as follows:

        with muninn.open(archive_name) as archive:
            product = archive.ingest(file_path)

    Please see the Muninn documentation for details about how to configure
    a Muninn archive (and also set an environment variable so Muninn can find
    its configuration file.)
    """

    @staticmethod
    def create(configuration, id=None):
        options = config.parse(configuration.get("archive", {}), _ArchiveConfig)
        _ArchiveConfig.validate(options)

        # Load and create the database backend.
        database_name = options.pop('database')
        database_module = _load_database_module(database_name)
        database = database_module.create(configuration)

        # Load and create the storage backend.
        storage_name = options.pop('storage')
        if storage_name == 'none':
            storage = None
        else:
            storage_module = _load_storage_module(storage_name)
            storage = storage_module.create(configuration, options.get('tempdir', None), options.get('auth_file', None))

        # Create the archive.
        namespace_extensions = options.pop("namespace_extensions", [])
        product_type_extensions = options.pop("product_type_extensions", [])
        remote_backend_extensions = options.pop("remote_backend_extensions", [])
        hook_extensions = options.pop("hook_extensions", [])
        archive = Archive(database=database, storage=storage, id=id,
                          configuration=configuration, **options)

        # Register core namespace.
        archive.register_namespace("core", Core)

        # Register custom extensions.
        for name in namespace_extensions:
            extension = _load_extension(name)
            _register_extension(archive, extension, name, 'namespace')

        for name in product_type_extensions:
            extension = _load_extension(name)
            _register_extension(archive, extension, name, 'product_type', get_name='product_type_plugin')

        for name in remote_backend_extensions:
            extension = _load_extension(name)
            _register_extension(archive, extension, name, 'remote_backend')

        for name in hook_extensions:
            extension = _load_extension(name)
            _register_extension(archive, extension, name, 'hook_extension')

        return archive

    def __init__(self, database, storage, cascade_grace_period=0,
                 max_cascade_cycles=25, auth_file=None, id=None, tempdir=None,
                 configuration=None):
        self._configuration = configuration

        self._cascade_grace_period = datetime.timedelta(minutes=cascade_grace_period)
        self._max_cascade_cycles = max_cascade_cycles
        self._auth_file = auth_file

        self._namespace_schemas = {}
        self._product_type_plugins = {}
        self._remote_backend_plugins = {}
        self._hook_extensions = collections.OrderedDict()

        self._export_formats = set()

        self._database = database
        self._database.initialize(self._namespace_schemas)

        self._storage = storage
        self._tempdir = tempdir
        self.id = id  # Archive id (usually name of configuration file)

        _register_extension(self, remote, 'muninn.remote', 'remote_backend')

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def register_namespace(self, namespace, schema):
        """Register a namespace. A valid namespace identifier starts
        with a lowercase character, and can contain any number of
        additional lowercase characters, underscores, or digits.

        Arguments:
        namespace -- Namespace identifier.
        schema    -- Schema definition of the namespace.
        """
        if not re.match(r"[a-z][_a-z0-9]*$", namespace):
            raise Error("invalid namespace identifier %s" % namespace)
        if namespace in self._namespace_schemas:
            raise Error("redefinition of namespace: \"%s\"" % namespace)

        self._namespace_schemas[namespace] = schema

    def namespace_schema(self, namespace):
        """Return the schema definition of the specified namespace.
        """
        try:
            return self._namespace_schemas[namespace]
        except KeyError:
            raise Error("unregistered namespace: \"%s\"; registered namespaces: %s" %
                        (namespace, util.quoted_list(self._namespace_schemas.keys())))

    def namespaces(self):
        """Return a list containing all registered namespaces.
        """
        return list(self._namespace_schemas.keys())

    def register_product_type(self, product_type, plugin):
        """Register a product type.

        Arguments:
        product_type -- Product type name
        plugin       -- Reference to an object that implements the product type
                        plugin API and as such takes care of the details of
                        extracting product properties from products of the
                        specified product type.
        """
        if product_type in self._product_type_plugins:
            raise Error("redefinition of product type: \"%s\"" % product_type)

        # Quick verify of the plugin interface
        for attr in ['use_enclosing_directory']:
            if not hasattr(plugin, attr):
                raise Error("missing '%s' attribute in plugin for product type \"%s\"" % (attr, product_type))

        if hasattr(plugin, 'use_hash') and not hasattr(plugin, 'hash_type'):
            warnings.warn("'use_hash' option is deprecated (use 'hash_type')")

        methods = ['identify', 'analyze', 'archive_path']
        if plugin.use_enclosing_directory:
            methods += ['enclosing_directory']
        for method in methods:
            if not hasattr(plugin, method):
                raise Error("missing '%s' method in plugin for product type \"%s\"" % (method, product_type))

        self._product_type_plugins[product_type] = plugin
        self._update_export_formats(plugin)

    def product_type_plugin(self, product_type):
        """Return a reference to the specified product type plugin.

        Arguments:
        product_type -- Product type name
        """
        try:
            return self._product_type_plugins[product_type]
        except KeyError:
            raise Error("undefined product type: \"%s\"; defined product types: %s" %
                        (product_type, util.quoted_list(self._product_type_plugins.keys())))

    def product_types(self):
        """Return a list of registered product types."""
        return list(self._product_type_plugins.keys())

    def register_remote_backend(self, remote_backend, plugin):
        """Register a remote backend.

        Arguments:
        remote_backend -- Remote backend name.
        plugin         -- Reference to an object that implements the remote
                          backend plugin API and as such takes care of the
                          details of extracting product properties from
                          products of the specified remote backend.

        """
        if remote_backend in self._remote_backend_plugins:
            raise Error("redefinition of remote backend: \"%s\"" % remote_backend)

        for method in ['identify', 'pull']:
            if not hasattr(plugin, method):
                raise Error("missing '%s' method in plugin for remote backend \"%s\"" % (method, remote_backend))

        self._remote_backend_plugins[remote_backend] = plugin

    def remote_backend(self, remote_backend):
        """Return a reference to the specified remote backend plugin.

        Arguments:
        remote_backend -- Remote backend name.
        """
        try:
            return self._remote_backend_plugins[remote_backend]
        except KeyError:
            raise Error("unregistered remote backend: \"%s\"; registered remote backends: %s" %
                        (remote_backend, util.quoted_list(self._remote_backend.keys())))

    def remote_backends(self):
        """Return a list of supported remote backends."""
        return list(self._remote_backend_plugins.keys())

    def register_hook_extension(self, hook_extension, plugin):
        """Register a hook extension.

        Arguments:
        hook_extension -- Hook extension name
        plugin         -- Reference to an object that implements the hook
                          extension plugin API
        """
        if hook_extension in self._hook_extensions:
            raise Error("redefinition of hook extension: \"%s\"" % hook_extension)

        self._hook_extensions[hook_extension] = plugin

    def hook_extension(self, hook_extension):
        """Return the hook extension with the specified name.

        Arguments:
        hook_extension -- Hook extension name
        """
        try:
            return self._hook_extensions[hook_extension]
        except KeyError:
            raise Error("unregistered hook extension: \"%s\"; registered hook extensions: %s" %
                        (hook_extension, util.quoted_list(self._hook_extensions.keys())))

    def hook_extensions(self):
        """Return a list of supported hook extensions."""
        return list(self._hook_extensions.keys())

    def _analyze_paths(self, plugin, paths):
        metadata = plugin.analyze(paths)

        if hasattr(plugin, 'namespaces'):
            namespaces = plugin.namespaces
        else:
            namespaces = []

        for namespace in metadata:
            if namespace != 'core' and namespace not in namespaces:
                warnings.warn("plugin.namespaces does not contain '%s'" % namespace, DeprecationWarning)

        if isinstance(metadata, (tuple, list)):
            properties, tags = metadata
        else:
            properties, tags = metadata, []

        return properties, tags

    def _check_paths(self, paths, action):
        if isinstance(paths, basestring):
            paths = [paths]

        if not paths:
            raise Error("nothing to %s" % action)

        # Use absolute paths to make error messages more useful, and to avoid broken links when ingesting/attaching
        # a product using symbolic links.
        paths = [os.path.realpath(path) for path in paths]

        # Ensure that the set of files and / or directories that make up the product does not contain duplicate
        # basenames.
        basenames = [os.path.basename(path) for path in paths]
        if len(set(basenames)) < len(basenames):
            raise Error("basename of each part should be unique for multi-part products")

        return [path.replace('\\', '/') for path in paths]

    def _extract_hash_type(self, hash_value):
        prefix, middle, _ = hash_value.partition(':')
        if middle == ':' and prefix in HASH_ALGORITHMS:
            return prefix

    def _get_product(self, uuid=None, namespaces=None, property_names=None, must_exist=True, **kwargs):
        if uuid is not None:
            kwargs['uuid'] = uuid

        if namespaces is None:
            namespaces = self.namespaces()

        cond = ' and '.join(['%s == @%s' % (key, key) for key in kwargs])

        products = self.search(cond, parameters=kwargs, namespaces=namespaces, property_names=property_names)

        if len(products) == 0:
            if must_exist:
                cond = ' and '.join(['%s=%s' % item for item in kwargs.items()])
                raise Error('No product found: %s' % cond)
            else:
                return None

        else:
            assert len(products) == 1
            return products[0]

    def _get_products(self, where, parameters=None, namespaces=None, property_names=None):  # TODO generator?
        if isinstance(where, basestring):
            return self.search(where, parameters=parameters, namespaces=namespaces, property_names=property_names)

        if isinstance(where, (UUID, Struct)):
            where = [where]
        else:
            try:
                where = list(where)
            except Exception:
                raise Error('Invalid product selection')

        products = []
        for term in where:
            if isinstance(term, UUID):
                product_uuid = term
            elif isinstance(term, Struct):
                product_uuid = term.core.uuid
            else:
                raise Error('Invalid product selection')
            products.append(self._get_product(product_uuid, namespaces=namespaces, property_names=property_names))

        return products

    def _plugin_hash_type(self, plugin):
        if hasattr(plugin, 'hash_type'):
            if not plugin.hash_type:
                return None
            else:
                return plugin.hash_type

        elif hasattr(plugin, 'use_hash'):
            if plugin.use_hash:
                return 'sha1'
            else:
                return None

        else:
            return 'md5'  # default after use_hash deprecation

    def _product_path(self, product):
        if self._storage is None or getattr(product.core, 'archive_path', None) is None:
            return None

        return os.path.join(product.core.archive_path, product.core.physical_name)

    def _purge(self, product):
        # get full product
        product = self._get_product(product.core.uuid)

        # Remove the product from the product catalogue.
        self._database.delete_product_properties(product.core.uuid)

        # Remove any data in storage associated with the product.
        self._remove_storage(product)

        # Run the post remove hook (if defined by the product type plugin or hook extensions).
        self._run_hooks('post_remove_hook', product, reverse=True)

    def _relocate(self, product, properties=None, paths=None):
        """Relocate a product to the archive_path reported by the product type plugin.
        Returns the new archive_path if the product was moved and optionally updated
        (local) product paths.
        """
        product_archive_path = product.core.archive_path
        if properties:
            product = copy.deepcopy(product)
            product.update(properties)
        plugin = self.product_type_plugin(product.core.product_type)
        plugin_archive_path = plugin.archive_path(product)

        if product_archive_path != plugin_archive_path:
            paths = self._storage.move(product, plugin_archive_path, paths)
            return plugin_archive_path, paths
        else:
            return None, paths

    def _remove_storage(self, product):
        # If the product has no data in storage associated with it, return.
        if self._storage is None:
            return
        product_path = self._product_path(product)
        if product_path is None:
            return

        # Remove the data associated with the product from storage.
        self._storage.delete(product_path, product)

    def _retrieve(self, product, target_path, use_symlinks=False, verify_hash=False):
        if 'archive_path' in product.core:
            # Determine the path of the product in storage.
            product_path = self._product_path(product)

            # Get the product type specific plugin.
            plugin = self.product_type_plugin(product.core.product_type)
            use_enclosing_directory = plugin.use_enclosing_directory

            # Symbolic link or copy the product at or to the specified target directory.
            paths = self._storage.get(product, target_path, use_enclosing_directory, use_symlinks)

            if verify_hash and 'hash' in product.core and not self._verify_hash(product, paths):
                raise Error("retrieved product '%s' (%s) has incorrect hash" %
                            (product.core.product_name, product.core.uuid))

            if not use_enclosing_directory:
                if len(paths) == 1:
                    paths = paths[0]

        elif 'remote_url' in product.core:
            retrieve_files = remote.retrieve_function(self, product, verify_hash)
            paths = retrieve_files(target_path)
            if len(paths) == 1:
                paths = paths[0]

        else:
            raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))

        return paths

    def _run_hooks(self, hook_name, properties, reverse=False, paths=None):
        plugins = list(self._hook_extensions.values())
        plugin = self._product_type_plugins.get(properties.core.product_type)
        if plugin is not None:
            plugins.insert(0, plugin)

        if reverse:
            plugins = reversed(plugins)

        for plugin in plugins:
            hook_method = getattr(plugin, hook_name, None)
            if hook_method is not None:
                if _inspect_nargs(hook_method) == 4:
                    hook_method(self, properties, paths)
                else:
                    hook_method(self, properties)

    def _run_for_product(self, product, fn, use_enclosing_directory):
        if self._storage is None:
            remote_url = product.core.remote_url

            if remote_url.startswith('file://'):
                product_path = product.core.remote_url[7:]
                if os.path.isdir(product_path):
                    paths = [os.path.join(product_path, basename) for basename in os.listdir(product_path)]
                else:
                    paths = [product_path]
                return fn(paths)

            else:
                with util.TemporaryDirectory(prefix=".run_for_product-",
                                             suffix="-%s" % product.core.uuid.hex) as tmp_path:
                    retrieve_files = remote.retrieve_function(self, product, True)  # TODO verify hash?
                    retrieve_files(tmp_path)
                    paths = [os.path.join(tmp_path, basename) for basename in os.listdir(tmp_path)]
                    return fn(paths)

        else:
            return self._storage.run_for_product(product, fn, use_enclosing_directory)

    def _strip(self, product):
        # Set the archive path to None to indicate the product has no data in storage associated with it.
        self.update_properties(Struct({'core': {'active': True, 'archive_path': None, 'archive_date': None}}),
                               product.core.uuid)

        # Remove any data in storage associated with the product.
        self._remove_storage(product)

    def _update_export_formats(self, plugin):
        # Find all callables of which the name starts with "export_". The remainder of the name is used as the name of
        # the export format.
        for export_method_name in dir(plugin):
            if not export_method_name.startswith("export_"):
                continue

            _, _, export_format = export_method_name.partition("_")
            if not export_format:
                continue

            export_method = getattr(plugin, export_method_name)
            if export_method is None or not hasattr(export_method, "__call__"):
                continue

            self._export_formats.add(export_format)

    def _update_metadata_date(self, properties):
        """ add a core.metadata_date field if it did not yet exist and set it to the current date """
        if "core" not in properties:
            properties.core = Struct()
        properties.core.metadata_date = self._database.server_time_utc()

    def _verify_hash(self, product, paths=None):
        if 'archive_path' in product.core or 'remote_url' in product.core:
            if 'hash' not in product.core:
                raise Error("no hash available for product '%s' (%s)" %
                            (product.core.product_name, product.core.uuid))

            stored_hash = product.core.hash
            hash_type = self._extract_hash_type(stored_hash)
            plugin = self.product_type_plugin(product.core.product_type)

            if hash_type is None:
                hash_type = 'sha1'  # default before use_hash deprecation
                stored_hash = 'sha1:' + stored_hash

            if paths is None:
                product_hash = functools.partial(util.product_hash, hash_type=hash_type)
                current_hash = self._run_for_product(product, product_hash, plugin.use_enclosing_directory)
            else:
                current_hash = util.product_hash(paths, hash_type=hash_type)

            if current_hash != stored_hash:
                return False

        return True

    def attach(self, paths, product_type=None, use_symlinks=None,
               verify_hash=False, verify_hash_before=False,
               use_current_path=False, force=False):
        """Add a product to the archive using an existing metadata record in the database.

        This function acts as the inverse of a strip(). A metadata record for this product should already exist in
        the database and no product should exist for it in the archive.

        The existing metadata record is found by performing a search based on product_type and physical_name.

        Arguments:
        paths            -- List of paths pointing to product files.
        product_type     -- Product type of the product to ingest. If left unspecified, an attempt will be made to
                            determine the product type automatically. By default, the product type will be determined
                            automatically.
        use_symlinks     -- If set to True, symbolic links to the original product will be stored in the archive
                            instead of a copy of the original product. If set to None, the value of the corresponding
                            archive wide configuration option will be used. By default, the archive configuration will
                            be used.
                            This option is ignored if use_current_path=True.
        verify_hash      -- If set to True then, after the ingestion, the product in the archive will be matched against
                            the hash from the metadata (only if the metadata contained a hash).
        verify_hash_before  --  If set to True then, before the product is attached to the archive, it will be matched
                            against the metadata hash (if it exists).
        use_current_path -- Ingest the product by keeping the file(s) at the current path (which must be inside the
                            root directory of the archive).
                            This option is ignored if `ingest_product` is False.
        force            -- If set to True, then skip default size check between product and existing metadata.

        Returns:
        The attached product.
        """
        if self._storage is None:
            raise Error('"attach" operation not available for storage=none')

        paths = self._check_paths(paths, 'attach')

        if product_type is None:
            product_type = self.identify(paths)
        plugin = self.product_type_plugin(product_type)

        properties, tags = self._analyze_paths(plugin, paths)

        # Determine physical product name
        if plugin.use_enclosing_directory:
            physical_name = plugin.enclosing_directory(properties)
        elif len(paths) == 1:
            physical_name = os.path.basename(paths[0])
        else:
            raise Error("cannot attach multi-file product without enclosing directory")

        # Find product in catalogue
        product = self._get_product(product_type=product_type, physical_name=physical_name,
                                    namespaces=getattr(plugin, 'namespaces', []))

        # Determine archive path
        if 'archive_path' in product.core:
            raise Error("product with physical_name '%s' is already in the archive" % physical_name)
        if use_current_path:
            archive_path = self._storage.current_archive_path(paths, product)
        else:
            archive_path = plugin.archive_path(product)

        # Check size match
        size = util.product_size(paths)
        if not force and size != product.core.size:
            raise Error("size mismatch between product and existing metadata")

        # Check hash match
        if verify_hash_before:
            stored_hash = getattr(product.core, 'hash', None)
            if stored_hash is not None:
                hash_type = self._extract_hash_type(stored_hash)
                if hash_type is None:
                    stored_hash = 'sha1:' + stored_hash
                    hash_type = 'sha1'
                product_hash = util.product_hash(paths, hash_type=hash_type)
                if product_hash != stored_hash:
                    raise Error("hash mismatch between product and existing metadata")

        # Determine hash using plugin hash type
        hash_type = self._plugin_hash_type(plugin)
        if hash_type is not None:
            product_hash = util.product_hash(paths, hash_type=hash_type)
        else:
            product_hash = None

        # Set properties and deactivate while we attach the product
        product.core.active = False
        product.core.size = size
        product.core.archive_path = archive_path
        metadata = {
            'active': False,
            'size': size,
            'archive_path': archive_path
        }
        if hash_type is not None:
            metadata['hash'] = product_hash
        self.update_properties(Struct({'core': metadata}), product.core.uuid)

        # Store the product into the archive.
        use_enclosing_directory = plugin.use_enclosing_directory
        try:
            self._storage.put(paths, product, use_enclosing_directory, use_symlinks)
        except Exception as e:
            if not (isinstance(e, StorageError) and e.anything_stored):
                # reset state to before attach
                metadata = {'active': True, 'archive_path': None}
                self.update_properties(Struct({'core': metadata}), product.core.uuid)
            if isinstance(e, StorageError):
                raise e.orig
            else:
                raise

        # Verify product hash after copy
        if verify_hash:
            if self.verify_hash(product.core.uuid):
                raise Error("ingested product has incorrect hash")

        # Activate product
        product.core.active = True
        product.core.archive_date = self._database.server_time_utc()
        metadata = {
            'active': True,
            'archive_date': product.core.archive_date,
        }
        self.update_properties(Struct({'core': metadata}), product.core.uuid)

        return product

    def auth_file(self):
        """Return the path of the authentication file to download from remote locations."""
        return self._auth_file

    def cleanup_derived_products(self):
        """Clean up all derived products for which the source products no
        longer exist, as specified by the cascade rule configured in the
        respective product type plugins.

        Please see the Muninn documentation for more information on how
        to configure cascade rules.
        """
        repeat = True
        cycle = 0
        while repeat and cycle < self._max_cascade_cycles:
            repeat = False
            cycle += 1
            for product_type in self.product_types():
                plugin = self.product_type_plugin(product_type)

                cascade_rule = getattr(plugin, "cascade_rule", CascadeRule.IGNORE)
                if cascade_rule == CascadeRule.IGNORE:
                    continue

                strip = cascade_rule in (CascadeRule.CASCADE_PURGE_AS_STRIP, CascadeRule.STRIP)
                products = self._database.find_products_without_source(product_type, self._cascade_grace_period, strip)
                if products:
                    repeat = True

                if strip:
                    for product in products:
                        self._strip(product)
                else:
                    for product in products:
                        self._purge(product)

                if cascade_rule in (CascadeRule.CASCADE_PURGE_AS_STRIP, CascadeRule.CASCADE_PURGE):
                    continue

                products = self._database.find_products_without_available_source(product_type)
                if products:
                    repeat = True

                if cascade_rule in (CascadeRule.STRIP, CascadeRule.CASCADE):
                    for product in products:
                        self._strip(product)
                else:
                    for product in products:
                        self._purge(product)

    def close(self):
        """Close the archive immediately instead of when (and if) the archive
        instance is collected.

        Using the archive after calling this function results in undefined behavior.
        """
        self._database.disconnect()

    def count(self, where="", parameters={}):
        """Return the number of products matching the specified search expression.

        Arguments:
        where       --  Search expression.
        parameters  --  Parameters referenced in the search expression (if any).
        """
        return self._database.count(where, parameters)

    def create_properties(self, properties, disable_hooks=False):
        """Create a record for the given product (as defined by the
        provided dictionary of properties) in the product catalogue.
        An important side effect of this operation is that it will
        fail if:

            1. The `core.uuid` is not unique within the product catalogue.
            2. The combination of `core.archive_path` and `core.physical_name` is
               not unique within the product catalogue.

        Arguments:
        properties    -- The product properties.
        disable_hooks -- Do not execute any hooks.

        """
        self._update_metadata_date(properties)
        self._database.insert_product_properties(properties)

        # Run the post create hook (if defined by the product type plugin or hook extensions).
        if not disable_hooks:
            self._run_hooks('post_create_hook', properties)

    def delete_properties(self, where="", parameters={}):
        """Remove properties for one or more products from the catalogue.

        This function will _not_ remove any product files from storage and
        will _not_ trigger any of the specific cascade rules.

        Arguments:
        where       --  Search expression or one or more product uuid(s) or properties.
        parameters  --  Parameters referenced in the search expression.

        Returns:
        The number of updated products
        """
        products = self._get_products(where, parameters, property_names=['uuid'])
        for product in products:
            self._database.delete_product_properties(product.core.uuid)
        return len(products)

    def derived_products(self, uuid):
        """Return the UUIDs of the products that are linked to the given
        product as derived products.

        Arguments:
        uuid -- Product UUID.
        """
        return self._database.derived_products(uuid)

    def destroy(self):
        """Completely remove the archive, including both the products and the
        product catalogue.

        Using the archive after calling this function results in undefined
        behavior. The prepare() function can be used to bring the archive back
        into a useable state.
        """
        self.destroy_catalogue()

        if self._storage is not None:
            self._storage.destroy()

    def destroy_catalogue(self):
        """Completely remove the catalogue database, but leave the datastore in storage untouched.

        Using the archive after calling this function results in undefined behavior.
        Using the prepare_catalogue() function and ingesting all products again, can bring the archive
        back into a useable state.
        """
        # Call the backend to remove anything related to the archive.
        if self._database.exists():
            self._database.destroy()

    def export(self, where="", parameters={}, target_path=os.path.curdir, format=None):
        """Export one or more products from the archive.

        By default, a copy of the original product will be retrieved from the archive. This default behavior can be
        customized by the product type plugin. For example, the custom implementation for a certain product type might
        retrieve one or more derived products and bundle them together with the product itself.

        Arguments:
        where           --  Search expression or one or more product uuid(s) or properties.
        parameters      --  Parameters referenced in the search expression.
        target_path     --  Directory in which the retrieved products will be stored.
        format          --  Format in which the products will be exported.

        Returns:
        Either a list containing the export paths for the
        exported products (when a search expression or multiple
        properties/UUIDs were passed), or a single export path.
        """
        export_method_name = "export"
        if format is not None:
            if re.match("[a-zA-Z]\\w*$", format) is None:
                raise Error("invalid export format '%s'" % format)
            export_method_name += "_" + format

        products = self._get_products(where, parameters, namespaces=self.namespaces())

        result = []
        for product in products:
            if not product.core.active:
                raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))

            # Use the format specific export_<format>() method from the product type plugin, or the default export()
            # method if no format was specified. Call self._retrieve() as a fall back if the product type plugin does
            # not define its own export() method.
            #
            # Note the use of getattr() / hasattr() instead of a try + except AttributeError block, to avoid hiding
            # AttributeError instances raised by the plugin.

            plugin = self.product_type_plugin(product.core.product_type)
            export_method = getattr(plugin, export_method_name, None)

            if export_method is not None:
                if _inspect_nargs(export_method) == 5:
                    def _export(paths):
                        exported_path = export_method(self, product, target_path, paths)
                        result.append(exported_path)
                    self._run_for_product(product, _export, plugin.use_enclosing_directory)
                else:
                    exported_path = export_method(self, product, target_path)
                    result.append(exported_path)

            elif format is not None:
                raise Error("export format '%s' not supported for product '%s' (%s)" %
                            (format, product.core.product_name, product.core.uuid))

            else:
                exported_path = self._retrieve(product, target_path, False)
                result.append(exported_path)

        if isinstance(where, (UUID, Struct)):
            return result[0]
        else:
            return result

    def export_formats(self):
        """Return a list of supported alternative export formats."""
        return list(self._export_formats)

    @staticmethod
    def generate_uuid():
        """Return a new generated UUID that can be used as UUID for a product metadata record."""
        return uuid4()

    def identify(self, paths):
        """Determine the product type of the product (specified as a single path, or a list of paths if it is a
        multi-part product).

        Arguments:
        paths            -- List of paths pointing to product files.

        Returns:
        The determined product type.
        """
        for product_type, plugin in self._product_type_plugins.items():
            if plugin.identify(paths):
                return product_type

        raise Error("unable to identify product: \"%s\"" % paths)

    def ingest(self, paths, product_type=None, properties=None, ingest_product=True, use_symlinks=None,
               verify_hash=False, use_current_path=False, force=False):
        """Ingest a product into the archive. Multiple paths can be specified, but the set of files and/or directories
        these paths refer to is always ingested as a single logical product.

        Product ingestion consists of two steps. First, product properties are extracted from the product and are used
        to create an entry for the product in the product catalogue. Second, the product itself is ingested, either by
        copying the product or by creating symbolic links to the product.

        If the product to be ingested is already located at the target location within the archive (and there was not
        already another catalogue entry pointing to it), muninn will leave the product at its location as-is, and won't
        try to copy/symlink it.

        Arguments:
        paths            -- List of paths pointing to product files.
        product_type     -- Product type of the product to ingest. If left unspecified, an attempt will be made to
                            determine the product type automatically. By default, the product type will be determined
                            automatically.
        properties       -- Used as product properties if specified. No properties will be extracted from the product
                            in this case.
        ingest_product   -- If set to False, the product itself will not be ingested into the archive, only its
                            properties. By default, the product will be ingested.
        use_symlinks     -- If set to True, symbolic links to the original product will be stored in the archive
                            instead of a copy of the original product. If set to None, the value of the corresponding
                            archive wide configuration option will be used. By default, the archive configuration will
                            be used.
                            This option is ignored if `use_current_path` is True.
        verify_hash      -- If set to True then, after the ingestion, the product in the archive will be matched against
                            the hash from the metadata (only if the metadata contained a hash).
        use_current_path -- Ingest the product by keeping the file(s) at the current path (which must be inside the
                            root directory of the archive).
                            This option is ignored if `ingest_product` is False.
        force            -- If set to True then any existing product with the same type and name (unique constraint)
                            will be removed before ingestion, including partially ingested products.
                            NB. Depending on product type specific cascade rules, removing a product can result in one
                            or more derived products being removed (or stripped) along with it.

        Returns:
        The ingested product.
        """
        paths = self._check_paths(paths, 'ingest')

        # Get the product type plugin.
        if product_type is None:
            product_type = self.identify(paths)
        plugin = self.product_type_plugin(product_type)

        # Extract product metadata.
        if properties is None:
            properties, tags = self._analyze_paths(plugin, paths)
        else:
            properties, tags = copy.deepcopy(properties), []

        assert properties is not None and "core" in properties
        assert "product_name" in properties.core and properties.core.product_name, \
            "product_name is required in core.properties"

        # Set core product properties that are not determined by the plugin.
        # Note that metadata_date is set automatically by create_properties()
        # and archive_date is properly set when we activate the product.
        properties.core.uuid = self.generate_uuid()
        properties.core.active = False
        properties.core.hash = None
        properties.core.size = util.product_size(paths)  # TODO determine after?
        properties.core.metadata_date = None
        properties.core.archive_date = None
        properties.core.archive_path = None
        properties.core.product_type = product_type
        properties.core.physical_name = None

        # Determine physical product name.
        if plugin.use_enclosing_directory:
            properties.core.physical_name = plugin.enclosing_directory(properties)
        elif len(paths) == 1:
            properties.core.physical_name = os.path.basename(paths[0])
        else:
            raise Error("cannot ingest multi-file product without enclosing directory")

        if self._storage is None:
            ingest_product = False

        # Determine archive path
        elif ingest_product:
            if use_current_path:
                properties.core.archive_path = self._storage.current_archive_path(paths, properties)
            else:
                properties.core.archive_path = plugin.archive_path(properties)

        # Remove existing product with the same product type and name before ingesting
        if force:
            existing = self._get_product(product_type=properties.core.product_type,
                                         product_name=properties.core.product_name,
                                         must_exist=False)
            if existing is not None:
                if 'archive_path' in existing.core and existing.core.archive_path is not None:
                    ingest_path = os.path.dirname(paths[0])
                    if plugin.use_enclosing_directory:
                        ingest_path = os.path.dirname(ingest_path)
                    current_path = self.root()
                    if existing.core.archive_path:
                        current_path = util.fwd_join(current_path, existing.core.archive_path)
                    if existing.core.archive_path != properties.core.archive_path:
                        raise Error('cannot force ingest because of archive_path mismatch')
                    if ingest_path == current_path:
                        # do not remove the product being ingested (only remove from catalogue)
                        self.delete_properties(existing.core.uuid)
                    else:
                        self.remove(existing.core.uuid, force=True)
                else:
                    self.delete_properties(existing.core.uuid)

        self.create_properties(properties, disable_hooks=True)

        # Try to determine the product hash and ingest the product into the archive.
        try:
            # Determine product hash. Since it is an expensive operation, the hash is computed after inserting the
            # product properties so we won't needlessly compute it for products that fail ingestion into the catalogue.
            hash_type = self._plugin_hash_type(plugin)
            if hash_type is not None:
                try:
                    properties.core.hash = util.product_hash(paths, hash_type=hash_type)
                except EnvironmentError as _error:
                    raise Error("cannot determine product hash [%s]" % (_error,))
                self.update_properties(Struct({'core': {'hash': properties.core.hash}}), properties.core.uuid)

            if ingest_product:
                use_enclosing_directory = plugin.use_enclosing_directory
                self._storage.put(paths, properties, use_enclosing_directory, use_symlinks)
                properties.core.archive_date = self._database.server_time_utc()

            elif self._storage is None:
                if len(paths) == 1:
                    properties.core.remote_url = 'file://' + os.path.realpath(paths[0])
                else:
                    properties.core.remote_url = 'file://' + os.path.realpath(os.path.dirname(paths[0]))  # TODO test
                properties.core.remote_url = properties.core.remote_url.replace('\\', '/')
                self.update_properties(Struct({'core': {'remote_url': properties.core.remote_url}}),
                                       properties.core.uuid)

        except Exception as e:
            if not (isinstance(e, StorageError) and e.anything_stored):
                # Try to remove the entry for this product from the product catalogue.
                self._database.delete_product_properties(properties.core.uuid)
            if isinstance(e, StorageError):
                raise e.orig
            else:
                raise

        # Update archive date and activate product
        properties.core.active = True
        metadata = {
            'active': properties.core.active,
            'archive_date': properties.core.archive_date,
        }
        self.update_properties(Struct({'core': metadata}), properties.core.uuid)

        # Set product tags.
        self._database.tag(properties.core.uuid, tags)

        # Verify product hash after copy
        if ingest_product and verify_hash:
            if self.verify_hash(properties.core.uuid):
                raise Error("ingested product has incorrect hash")

        # Run post create/ingest hooks (if defined by the product type plugin or hook extensions).
        if not ingest_product:
            self._run_hooks('post_create_hook', properties)
        else:
            self._run_hooks('post_ingest_hook', properties, paths=paths)

        return properties

    def link(self, uuid, source_uuids):
        """Link a product to one or more source products.

        Arguments:
        uuid         -- Product UUID.
        source_uuids -- Source UUIDs.
        """
        if isinstance(source_uuids, UUID):
            source_uuids = [source_uuids]

        self._database.link(uuid, source_uuids)

    def prepare(self, force=False):
        """Prepare the archive for (first) use.

        The root path will be created and the product catalogue will be
        initialized such that the archive is ready for use.

        Arguments:
        force   --  If set to True then any existing products and / or product
                    catalogue will be removed.
        """
        if not force:
            if self._storage is not None and self._storage.exists():
                raise Error("storage already exists")
            if self._database.exists():
                raise Error("database already exists")

        # Remove anything related to the archive.
        self.destroy()

        # Prepare the archive for use.
        self._database.prepare()
        if self._storage is not None:
            self._storage.prepare()

    def prepare_catalogue(self, dry_run=False):
        """Prepare the catalogue of the archive for (first) use.

        Arguments:
        dry_run -- Do not actually execute the preparation commands.

        Returns:
        The list of SQL commands that (would) have been executed by this function.
        """
        return self._database.prepare(dry_run=dry_run)

    def product_path(self, uuid_or_properties):
        """Return the path in storage to the specified product.

        Arguments:
        uuid_or_properties: UUID or dictionary of product properties.
        """
        if isinstance(uuid_or_properties, Struct):
            product = uuid_or_properties
        else:
            property_names = ['archive_path', 'physical_name']
            if isinstance(uuid_or_properties, UUID):
                product = self._get_product(uuid_or_properties,
                                            property_names=property_names)
            else:
                product = self._get_product(uuid_or_properties.core.uuid,
                                            property_names=property_names)

        archive_path = getattr(product.core, 'archive_path', None)
        remote_url = getattr(product.core, 'remote_url', None)
        if archive_path is not None and self._storage is not None:
            product_path = self._product_path(product)
            return os.path.join(self._storage.global_prefix, product_path)
        elif remote_url is not None:
            return remote_url

    def pull(self, where="", parameters={}, verify_hash=False, verify_hash_download=False):
        """Pull one or more remote products into the archive.

        Products should have a valid remote_url core metadata field and they should not yet exist in the local
        archive (i.e. the archive_path core metadata field should not be set).

        Arguments:
        where         --  Search expression or one or more product uuid(s) or properties.
        parameters    --  Parameters referenced in the search expression (if any).
        verify_hash   --  If set to True then, after the pull, the product in the archive will be matched against
                          the hash from the metadata (only if the metadata contained a hash).
        verify_hash_download  --  If set to True then, before the product is stored in the archive, the pulled
                          product will be matched against the metadata hash (if it exists).

        Returns:
        The number of pulled products.
        """

        if self._storage is None:
            raise Error('"pull" operation not available for storage=none')

        products = self._get_products(where, parameters, namespaces=self.namespaces())

        for product in products:
            if not product.core.active:
                raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))
            if 'archive_path' in product.core:
                raise Error("product '%s' (%s) is already in the local archive" %
                            (product.core.product_name, product.core.uuid))
            if 'remote_url' not in product.core:
                raise Error("product '%s' (%s) does not have a remote_url" %
                            (product.core.product_name, product.core.uuid))

            plugin = self.product_type_plugin(product.core.product_type)
            product.core.archive_path = plugin.archive_path(product)
            use_enclosing_directory = plugin.use_enclosing_directory

            # set archive_path and deactivate while we pull it in
            metadata = {'active': False, 'archive_path': product.core.archive_path}
            self.update_properties(Struct({'core': metadata}), product.core.uuid)

            def _pull(paths):
                if len(paths) > 1 and not use_enclosing_directory:
                    raise Error("cannot pull multi-file product without enclosing directory")

                # reactivate and update archive_date, size
                product_path = self._product_path(product)
                size = self._storage.size(product_path)
                metadata = {'active': True, 'archive_date': self._database.server_time_utc(), 'size': size}
                if 'hash' not in product.core or product.core.hash is None:
                    hash_type = self._plugin_hash_type(plugin)
                    if hash_type is not None:
                        try:
                            hash = util.product_hash(paths, hash_type=hash_type)
                        except EnvironmentError as _error:
                            raise Error("cannot determine product hash [%s]" % (_error,))
                        metadata['hash'] = hash
                self.update_properties(Struct({'core': metadata}), product.core.uuid)

                # verify product hash.
                if verify_hash and 'hash' in product.core:
                    if self.verify_hash(product.core.uuid):
                        raise Error("pulled product '%s' (%s) has incorrect hash" %
                                    (product.core.product_name, product.core.uuid))

                # Run the post pull hook (if defined by the product type plugin or hook extensions).
                self._run_hooks('post_pull_hook', product, paths=paths)

            # pull product
            try:
                retrieve_files = remote.retrieve_function(self, product, verify_hash_download)
                self._storage.put(None, product, use_enclosing_directory, use_symlinks=False,
                                  retrieve_files=retrieve_files, run_for_product=_pull)
            except Exception as e:
                if not (isinstance(e, StorageError) and e.anything_stored):
                    # reset state to before pull
                    metadata = {'active': True, 'archive_path': None, 'archive_date': None}
                    self.update_properties(Struct({'core': metadata}), product.core.uuid)
                if isinstance(e, StorageError):
                    raise e.orig
                else:
                    raise

        return len(products)

    def rebuild_properties(self, uuid, disable_hooks=False, use_current_path=False):
        """Rebuild product properties by re-extracting these properties (using product type plugins) from the
        products stored in the archive.
        Only properties and tags that are returned by the product type plugin will be updated. Other properties or
        tags will remain as they were.

        Arguments:
        uuid             -- Product UUID
        disable_hooks    -- Disable product type hooks (not meant for routine operation).
        use_current_path -- Do not attempt to relocate the product to the location specified in the product
                            type plugin. Useful for read-only archives.

        """
        restricted_properties = set(["uuid", "active", "hash", "size", "metadata_date", "archive_date", "archive_path",
                                     "product_type", "physical_name"])

        product = self._get_product(uuid)
        if not (product.core.active and ('archive_path' in product.core or 'remote_url' in product.core)):
            raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))

        # Extract product metadata.
        plugin = self.product_type_plugin(product.core.product_type)
        use_enclosing_directory = plugin.use_enclosing_directory

        def _rebuild_properties(paths):
            properties, tags = self._analyze_paths(plugin, paths)

            # Remove properties that should not be changed.
            assert "core" in properties
            for name in restricted_properties:
                try:
                    delattr(properties.core, name)
                except AttributeError:
                    pass

            # update size
            properties.core.size = util.product_size(paths)

            # Make sure product is stored in the correct location
            if not use_current_path and 'archive_path' in product.core:
                new_archive_path, paths = self._relocate(product, properties, paths)
                if new_archive_path is not None:
                    metadata = {'archive_path': new_archive_path}
                    self.update_properties(Struct({'core': metadata}), product.core.uuid)
                    properties.core.archive_path = new_archive_path

            # if product type has disabled hashing, remove existing hash values
            stored_hash = getattr(product.core, 'hash', None)
            plugin_hash_type = self._plugin_hash_type(plugin)

            if plugin_hash_type is None:
                if stored_hash is not None:
                    properties.core.hash = None

            # if product type has different hash algorithm, update hash values
            else:
                if stored_hash is None:
                    properties.core.hash = util.product_hash(paths, hash_type=plugin_hash_type)
                else:
                    hash_type = self._extract_hash_type(stored_hash)
                    if hash_type is None and plugin_hash_type == 'sha1':
                        properties.core.hash = plugin_hash_type + ':' + stored_hash
                    elif hash_type != plugin_hash_type:
                        properties.core.hash = util.product_hash(paths, hash_type=plugin_hash_type)

            # Update product properties.
            self.update_properties(properties, uuid=product.core.uuid, create_namespaces=True)

            # Update tags.
            self.tag(product.core.uuid, tags)

            # Run the post ingest hook (if defined by the product type plugin or hook extensions).
            if not disable_hooks:
                product.update(properties)
                if 'hash' not in product.core:
                    product.core.hash = None
                self._run_hooks('post_ingest_hook', product, paths=paths)

        self._run_for_product(product, _rebuild_properties, use_enclosing_directory)

    def rebuild_pull_properties(self, uuid, verify_hash=False, disable_hooks=False, use_current_path=False):
        """Refresh products by re-running the pull, but using the existing products stored in the archive.

        Arguments:
        uuid             -- Product UUID
        verify_hash      -- If set to True then the product in the archive will be matched against
                            the hash from the metadata (only if the metadata contained a hash).
        disable_hooks    -- Disable product type hooks (not meant for routine operation).
        use_current_path -- Do not attempt to relocate the product to the location specified in the product
                            type plugin. Useful for read-only archives.
        """
        if self._storage is None:
            raise Error('"rebuild_pull_properties" operation not available for storage=none')

        product = self._get_product(uuid)
        if 'archive_path' not in product.core:
            raise Error("cannot update missing product")
        if 'remote_url' not in product.core:
            raise Error("cannot pull products that have no remote_url")

        plugin = self.product_type_plugin(product.core.product_type)
        use_enclosing_directory = plugin.use_enclosing_directory

        # make sure product is stored in the correct location
        if not use_current_path:
            new_archive_path, _ = self._relocate(product)
            if new_archive_path is not None:
                metadata = {'archive_path': new_archive_path}
                self.update_properties(Struct({'core': metadata}), product.core.uuid)
                product.core.archive_path = new_archive_path

        # update size
        product_path = self._product_path(product)
        product.core.size = self._storage.size(product_path)

        def _rebuild_pull_properties(paths):
            # verify product hash.
            if verify_hash and 'hash' in product.core:
                if not self._verify_hash(product, paths):
                    raise Error("pulled product '%s' (%s) has incorrect hash" %
                                (product.core.product_name, product.core.uuid))

            # Run the post pull hook (if defined by the product type plugin or hook extensions).
            if not disable_hooks:
                self._run_hooks('post_pull_hook', product, paths=paths)

        self._run_for_product(product, _rebuild_pull_properties, use_enclosing_directory)

    def remove(self, where="", parameters={}, force=False, cascade=True):
        """Remove one or more products from the archive, both from storage as well as from the product catalogue.
        Return the number of products removed.

        NB. Depending on product type specific cascade rules, removing a product can result in one or more derived
        products being removed (or stripped) along with it. Such products are _not_ included in the returned count.

        Arguments:
        where       --  Search expression or one or more product uuid(s) or properties.
        parameters  --  Parameters referenced in the search expression.
        force       --  If set to True, also remove partially ingested products. This affects products for which a
                        failure occured during ingestion, as well as products in the process of being ingested. Use
                        this option with care.
        cascade     --  Apply cascade rules to strip/remove dependent products.

        Returns:
        The number of removed products.
        """
        products = self._get_products(where, parameters, property_names=_CORE_PROP_NAMES)

        for product in products:
            if not product.core.active and not force:
                raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))

            self._purge(product)

        # Remove (or strip) derived products if necessary.
        if cascade and len(products) > 0:
            self.cleanup_derived_products()

        return len(products)

    def retrieve(self, where="", parameters={}, target_path=os.path.curdir, use_symlinks=False, verify_hash=False):
        """Retrieve one or more products from the archive.

        Arguments:
        where           --  Search expression or one or more product uuid(s) or properties.
        parameters      --  Parameters referenced in the search expression.
        target_path     --  Directory under which the retrieved products will be stored.
        use_symlinks    --  If set to True, products will be retrieved as symbolic links to the original products kept
                            in the archive. If set to False, products will retrieved as copies of the original products.
                            By default, products will be retrieved as copies.
        verify_hash     --  If set to True then, after the retrieval, the product will be matched against the hash
                            from the metadata (only if the metadata contained a hash).

        Returns:
        Either a list containing the target paths for the
        retrieved products (when a search expression or multiple
        properties/uuids were passed), or a single target path.

        """
        products = self._get_products(where, parameters, property_names=_CORE_PROP_NAMES + ['remote_url'])

        result = []
        for product in products:
            if product.core.active and ('archive_path' in product.core or 'remote_url' in product.core):
                result.append(self._retrieve(product, target_path, use_symlinks, verify_hash))
            else:
                raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))

        if isinstance(where, (UUID, Struct)):
            return result[0]
        else:
            return result

    def retrieve_properties(self, uuid, namespaces=[], property_names=[]):
        """Return properties for the specified product.

        Arguments:
        uuid        -- Product UUID
        namespaces  -- List of namespaces of which the properties should be retrieved. By default, only properties
                       defined in the `core` namespace will be retrieved.
        """
        return self._get_product(uuid, namespaces=namespaces, property_names=property_names)

    def root(self):
        """Return the archive root path."""
        if self._storage is not None:
            return self._storage._root.replace('\\', '/')

    def search(self, where="", order_by=[], limit=None, parameters={}, namespaces=[], property_names=[]):
        """Search the product catalogue for products matching the specified search expression.

        Arguments:
        where       --  Search expression.
        order_by    --  A list of property names that determines the ordering of the results. If the list is empty, the
                        order of the results in undetermined and can very between calls to this function. Each property
                        name in this list can be provided with a '+' or '-' prefix, or without a prefix. A '+' prefix,
                        or no prefix denotes ascending sort order, a '-' prefix denotes decending sort order.
        limit       --  Limit the maximum number of results to the specified number.
        parameters  --  Parameters referenced in the search expression.
        namespaces  --  List of namespaces of which the properties should be retrieved. By default, only properties
                        defined in the `core` namespace will be retrieved.
        property_names  --  List of property names that should be returned. By default all properties of the `core`
                        namespace and those of the namespaces in the namespaces argument are included.
                        If this parameter is a non-empty list then only the referenced properties will be returned.
                        Properties are specified as `<namespace>.<identifier>`
                        (the namespace can be omitted for the `core` namespace).
                        If the `property_names` parameter is provided then the namespaces parameter is ignored.

        Returns:
        A list of matching products.
        """
        return self._database.search(where, order_by, limit, parameters, namespaces, property_names)

    def source_products(self, uuid):
        """Return the UUIDs of the products that are linked to the given product as source products.

        Arguments:
        uuid -- Product UUID
        """
        return self._database.source_products(uuid)

    def strip(self, where="", parameters={}, force=False, cascade=True):
        """Remove one or more products from storage only (not from the product catalogue).

        NB. Depending on product type specific cascade rules, stripping a product can result in one or more derived
        products being stripped (or removed) along with it.

        Arguments:
        where       --  Search expression or one or more product uuid(s) or properties.
        parameters  --  Parameters referenced in the search expression.
        force       --  If set to True, also strip partially ingested products. This affects products for which a
                        failure occured during ingestion, as well as products in the process of being ingested. Use
                        this option with care.
        cascade     --  Apply cascade rules to strip/purge dependent products.

        Returns:
        The number of stripped products.
        """
        if self._storage is None:
            raise Error('"strip" operation not available for storage=none')

        products = self._get_products(where, parameters, property_names=_CORE_PROP_NAMES)

        for product in products:
            if force:
                if product.core.active and 'archive_path' not in product.core and 'archive_date' not in product.core:
                    continue
            else:
                if 'archive_path' not in product.core:
                    continue
                if not product.core.active:
                    raise Error("product '%s' (%s) not available" % (product.core.product_name, product.core.uuid))

            self._strip(product)

        # Strip (or remove) derived products if necessary.
        if cascade and len(products) > 0:
            self.cleanup_derived_products()

        return len(products)

    def summary(self, where="", parameters=None, aggregates=None, group_by=None, group_by_tag=False,
                having=None, order_by=None):
        """Return a summary of the products matching the specified search expression.

        Arguments:
        where         --  Search expression.
        parameters    --  Parameters referenced in the search expression.
        aggregates    --  A list of property aggregates defined as `<property_name>.<reduce_fn>`.
                          Properties need to be of type long, integer, real, text or timestamp.
                          The reduce function can be `min`, `max`, `sum`, or `avg`.
                          `sum` and `avg` are not possible for text and timestamp properties.
                          A special property `validity_duration` (defined as `validity_stop` - `validity_start`)
                          can also be used.
        group_by      --  A list of property names whose values are used for grouping the aggregation results.
                          There will be a separate result row for each combination of `group_by` property values.
                          Properties need to be of type long, integer, boolean, text or timestamp.
                          Timestamps require a binning subscript which can be `year`, `month`, `yearmonth`, or `date`
                          (e.g. `validity_start.yearmonth`).
        group_by_tag  --  If set to True, results will also be grouped by available tag values.
                          Note that products will be counted multiple times if they have multiple tags
        having        --  A list of property aggregates defined as `<property_name>.<reduce_fn>`; properties need to be
                          of type long, integer, real, text or timestamp; the reduce function can be `min`, `max`,
                          `sum`, or `avg`; `sum` and `avg` are not possible for text and timestamp properties; a
                          special property `validity_duration` (defined as `validity_stop` - `validity_start`) can also
                          be used.
        order_by      --  A list of result column names that determines the ordering of the results. If the list is
                          empty, the order of the results is ordered by the `group_by` specification. Each name in the
                          list can have a `+` (ascending) or `-` (descending) prefix, or no prefix (ascending).

        Returns:
        A list of row tuples matching the search expression created from the arguments.
        """
        return self._database.summary(where, parameters, aggregates, group_by, group_by_tag, having, order_by)

    def tag(self, where=None, tags=None, parameters={}):
        """Set one or more tags on one or more product(s).

        Arguments:
        where       --  Search expression or one or more product uuid(s) or properties.
        tags        --  One or more tags.
        parameters  --  Parameters referenced in the search expression.
        """
        if isinstance(tags, basestring):
            tags = [tags]
        for tag in tags:
            if not isinstance(tag, basestring):
                raise Error('tag must be a string')

        if isinstance(where, UUID):
            self._database.tag(where, tags)
        else:
            products = self._get_products(where, parameters, property_names=['uuid'])
            for product in products:
                self._database.tag(product.core.uuid, tags)

    def tags(self, uuid):
        """Return the tags of a product.

        Arguments:
        uuid -- Product UUID.
        """
        return self._database.tags(uuid)

    def unlink(self, uuid, source_uuids=None):
        """Remove the link between a product and one or more of its source products.

        Arguments:
        uuid         -- Product UUID
        source_uuids -- Source product UUIDs
        """
        if isinstance(source_uuids, UUID):
            source_uuids = [source_uuids]

        self._database.unlink(uuid, source_uuids)

    def untag(self, where=None, tags=None, parameters={}):
        """Remove one or more tags from one or more product(s).

        Arguments:
        where       --  Search expression or one or more product uuid(s) or properties.
        tags        --  One or more tags (default: None, meaning all existing tags)
        parameters  --  Parameters referenced in the search expression.
        """
        if isinstance(tags, basestring):
            tags = [tags]

        if isinstance(where, UUID):
            self._database.untag(where, tags)
        else:
            products = self._get_products(where, parameters, property_names=['uuid'])
            for product in products:
                self._database.untag(product.core.uuid, tags)

    def update_properties(self, properties, uuid=None, create_namespaces=False):
        """Update product properties in the product catalogue. The UUID of the product to update will be taken from the
        `core.uuid` property if it is present in the specified properties. Otherwise, the UUID should be provided
        separately.

        This function allows any property to be changed with the exception of the product UUID, and therefore needs to
        be used with care. The recommended way to update product properties is to first retrieve them using either
        `retrieve_properties()` or `search()`, change the properties, and then use this function to update the product
        catalogue.

        Arguments:
        properties         -- Product properties
        uuid               -- UUID of the product to update. By default, the UUID will be taken from the `core.uuid`
                              property.
        create_namespaces  -- Test if all namespaces are already defined for the product, and create them if needed.
        """
        if create_namespaces:
            if 'core' in properties and 'uuid' in properties.core:
                uuid = properties.core.uuid if uuid is None else uuid
                if uuid != properties.core.uuid:
                    raise Error("specified uuid does not match uuid included in the specified product properties")

            existing_product = self._get_product(uuid)
            new_namespaces = list(set(vars(properties)) - set(vars(existing_product)))
        else:
            new_namespaces = None
        self._update_metadata_date(properties)
        self._database.update_product_properties(properties, uuid=uuid, new_namespaces=new_namespaces)

    def verify_hash(self, where="", parameters={}):
        """Verify the hash for one or more products in the archive.

        Products that are not active or are not in the archive will be skipped.
        If there is no hash available in the metadata for a product then an
        error will be raised.

        Arguments:
        where           --  Search expression or one or more product uuid(s) or properties.
        parameters      --  Parameters referenced in the search expression (if any).

        Returns:
        A list of UUIDs of products for which the verification failed.
        """
        property_names = _CORE_PROP_NAMES + ['hash', 'remote_url']
        products = self._get_products(where, parameters, property_names=property_names)

        failed_products = []
        for product in products:
            if not self._verify_hash(product):
                failed_products.append(product.core.uuid)

        return failed_products  # TODO different type of return value when single product passed?



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import logging
import sys
import os

import muninn

from muninn.tools.utils import Processor, create_parser, parse_args_and_run
from muninn.tools.ingest import CheckProductListAction, filter_paths, get_path_expansion_function


class AttachProcessor(Processor):

    def __init__(self, args):
        super(AttachProcessor, self).__init__(args)
        self.path_expansion_function = get_path_expansion_function(args.path_is_stem, args.path_is_enclosing_directory)
        self.use_symlinks = args.link
        self.force = args.force
        self.verify_hash = args.verify_hash
        self.verify_hash_before = args.verify_hash_before
        self.exclude = args.exclude
        self.product_type = args.product_type
        self.keep = args.keep

    def perform_operation(self, archive, path):
        path = os.path.abspath(path.strip())

        # Expand path into multiple files and/or directories that belong to the same product.
        try:
            product_paths = self.path_expansion_function(path)
        except muninn.Error as error:
            logging.error("%s: unable to determine which files or directories belong to product [%s]" % (path, error))
            return 0

        # Discard paths matching any of the user supplied exclude patterns.
        if self.exclude:
            product_paths = filter_paths(product_paths, self.exclude)

        if not product_paths:
            logging.error("%s: path does not match any files or directories" % path)
            return 0

        try:
            archive.attach(product_paths, self.product_type, use_symlinks=self.use_symlinks,
                           verify_hash=self.verify_hash, verify_hash_before=self.verify_hash_before,
                           use_current_path=self.keep, force=self.force)
        except muninn.Error as error:
            logging.error("%s: unable to attach product [%s]" % (path, error))
            return 0

        return 1


def attach(args):
    processor = AttachProcessor(args)
    with muninn.open(args.archive) as archive:
        if "-" in args.path:
            paths = [path for path in sys.stdin]
        else:
            paths = args.path
        return processor.process(archive, args, paths)


def main():
    parser = create_parser(description="Attach product to existing metadata entry in a muninn archive.", parallel=True)
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-s", "--path-is-stem", action="store_true", help="each product path is interpreted as a "
                       "stem; any file or directory of which the name starts with this stem is considered to be part "
                       "of the product")
    group.add_argument("-d", "--path-is-enclosing-directory", action="store_true", help="each product path is "
                       "interpreted as an enclosing directory; the actual product consists of any file or directory "
                       "found inside this enclosing directory; the enclosing directory itself is not considered part "
                       "of the product")
    parser.add_argument("-e", "--exclude", metavar="PATTERN", action="append", help="exclude any files or "
                        "directories whose basename matches PATTERN; *, ?, and [] can be used as wildcards; to match "
                        "a wildcard character literally, include it within brackets, e.g. [?] to match the character ? "
                        "literally")
    parser.add_argument("-t", "--product-type", help="force the product type of products to attach")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-l", "--link", action="store_true", help="attach symbolic links to each product")
    group.add_argument("-k", "--keep", action="store_true", help="attach product, using the current product path if it "
                                                                 "is in the muninn path, otherwise throws an error")
    parser.add_argument("-f", "--force", action="store_true", help="skip matching size check before attaching products")
    parser.add_argument("--verify-hash", action="store_true",
                        help="verify the hash of the product after it has been put in the archive")
    parser.add_argument("--verify-hash-before", action="store_true",
                        help="verify the hash of the product before it is put in the archive")
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    parser.add_argument("path", metavar="PATH", nargs="+", action=CheckProductListAction,
                        help="products to attach, or \"-\" to read the list of products from standard input")
    return parse_args_and_run(parser, attach)


if __name__ == '__main__':
    main()



import os

from .base import StorageBackend

from muninn.schema import Mapping, Text, Boolean
import muninn.util as util
from muninn.exceptions import Error, StorageError
import muninn.config as config


class _FSConfig(Mapping):
    _alias = "fs"

    root = Text()
    use_symlinks = Boolean(optional=True)


def create(configuration, tempdir, auth_file):
    options = config.parse(configuration.get("fs", {}), _FSConfig)
    _FSConfig.validate(options)
    return FilesystemStorageBackend(tempdir=tempdir, **options)


class FilesystemStorageBackend(StorageBackend):
    def __init__(self, root, use_symlinks=None, tempdir=None):
        super(FilesystemStorageBackend, self).__init__(tempdir)

        self.global_prefix = root
        self._root = os.path.realpath(root)
        self._use_symlinks = use_symlinks or False
        self.supports_symlinks = True

    def prepare(self):
        # Create the archive root path.
        try:
            util.make_path(self._root)
        except EnvironmentError as _error:
            raise Error("unable to create archive root path '%s' [%s]" % (self._root, _error))

    # tempdirs must be on the same file system for moves (below) to be atomic!
    def get_tmp_root(self, product):
        tmp_root = os.path.join(self._root, product.core.archive_path)
        util.make_path(tmp_root)
        return tmp_root

    def run_for_product(self, product, fn, use_enclosing_directory):
        product_path = os.path.join(self._root, product.core.archive_path, product.core.physical_name)
        if use_enclosing_directory:
            paths = [os.path.join(product_path, basename) for basename in os.listdir(product_path)]
        else:
            paths = [product_path]
        return fn(paths)

    def exists(self):
        return os.path.isdir(self._root)

    def destroy(self):
        if self.exists():
            try:
                util.remove_path(self._root)
            except EnvironmentError as _error:
                raise Error("unable to remove archive root path '%s' [%s]" % (self._root, _error))

    def current_archive_path(self, paths, properties):
        for path in paths:
            if not util.is_sub_path(os.path.realpath(path), self._root, allow_equal=True):
                raise Error("cannot ingest a file in-place if it is not inside the muninn archive root")

        abs_archive_path = os.path.dirname(os.path.realpath(paths[0]))

        if len(paths) > 1:
            # check whether all files have the right enclosing directory
            for path in paths:
                enclosing_directory = os.path.basename(os.path.dirname(os.path.realpath(path)))
                if enclosing_directory != properties.core.physical_name:
                    raise Error("multi-part product has invalid enclosing directory for in-place ingestion")
            abs_archive_path = os.path.dirname(abs_archive_path)

        # strip archive root
        return os.path.relpath(abs_archive_path, start=os.path.realpath(self._root))

    def put(self, paths, properties, use_enclosing_directory, use_symlinks=None,
            retrieve_files=None, run_for_product=None):

        if use_symlinks is None:
            use_symlinks = self._use_symlinks

        physical_name = properties.core.physical_name
        archive_path = properties.core.archive_path
        uuid = properties.core.uuid

        abs_archive_path = os.path.realpath(os.path.join(self._root, archive_path))
        abs_product_path = os.path.join(abs_archive_path, physical_name)

        # TODO separate this out like 'current_archive_path'
        if paths is not None and util.is_sub_path(os.path.realpath(paths[0]), abs_product_path, allow_equal=True):
            # Product should already be in the target location
            for path in paths:
                if not os.path.exists(path):
                    raise Error("product source path does not exist '%s'" % (path,))
                if not util.is_sub_path(os.path.realpath(path), abs_product_path, allow_equal=True):
                    raise Error("cannot ingest product where only part of the files are already at the "
                                "destination location")
        else:
            # Create destination location for product
            try:
                util.make_path(abs_archive_path)
            except EnvironmentError as _error:
                raise Error("cannot create parent destination path '%s' [%s]" % (abs_archive_path, _error))

            anything_stored = False

            # Create a temporary directory and transfer the product there, then move the product to its
            # destination within the archive.
            try:
                tmp_root = self.get_tmp_root(properties)
                with util.TemporaryDirectory(prefix=".put-", suffix="-%s" % uuid.hex,
                                             dir=tmp_root) as tmp_path:
                    try:
                        # Create enclosing directory if required.
                        if use_enclosing_directory:
                            tmp_path = os.path.join(tmp_path, physical_name)
                            util.make_path(tmp_path)

                        # Transfer the product (parts).
                        if retrieve_files:
                            # Retrieve product (parts).
                            paths = retrieve_files(tmp_path)
                        else:
                            if use_symlinks:
                                # Create symbolic link(s) for the product (parts).
                                if use_enclosing_directory:
                                    abs_path = abs_product_path
                                else:
                                    abs_path = abs_archive_path

                                for path in paths:
                                    if util.is_sub_path(path, self._root):
                                        # Create a relative symbolic link when the target is part of the archive
                                        # (i.e. when creating an intra-archive symbolic link). This ensures the
                                        # archive can be relocated without breaking intra-archive symbolic links.
                                        os.symlink(os.path.relpath(path, abs_path),
                                                   os.path.join(tmp_path, os.path.basename(path)))
                                    else:
                                        os.symlink(path, os.path.join(tmp_path, os.path.basename(path)))
                            else:
                                # Copy product (parts).
                                for path in paths:
                                    util.copy_path(path, tmp_path, resolve_root=True)

                        # Move the transferred product into its destination within the archive.
                        if use_enclosing_directory:
                            os.rename(tmp_path, abs_product_path)
                        else:
                            assert len(paths) == 1 and os.path.basename(paths[0]) == physical_name
                            tmp_product_path = os.path.join(tmp_path, physical_name)
                            os.rename(tmp_product_path, abs_product_path)
                        anything_stored = True
                    except EnvironmentError as _error:
                        raise Error("unable to transfer product to destination path '%s' [%s]" %
                                    (abs_product_path, _error))

                    # Run optional function on result
                    if run_for_product is not None:
                        self.run_for_product(properties, run_for_product, use_enclosing_directory)

            except Exception as e:
                raise StorageError(e, anything_stored)

    def get(self, product, target_path, use_enclosing_directory, use_symlinks=None):
        paths = []

        if use_symlinks is None:
            use_symlinks = self._use_symlinks

        product_path = os.path.join(self._root, product.core.archive_path, product.core.physical_name)
        product_basename = os.path.basename(product_path)

        try:
            if use_symlinks:
                if use_enclosing_directory:
                    for basename in os.listdir(product_path):
                        os.symlink(os.path.join(product_path, basename), os.path.join(target_path, basename))
                        paths.append(os.path.join(target_path, basename))
                else:
                    os.symlink(product_path, os.path.join(target_path, product_basename))
                    paths.append(os.path.join(target_path, product_basename))
            else:
                if use_enclosing_directory:
                    for basename in os.listdir(product_path):
                        util.copy_path(os.path.join(product_path, basename), target_path, resolve_root=True)
                        paths.append(os.path.join(target_path, basename))
                else:
                    util.copy_path(product_path, target_path, resolve_root=True)
                    paths.append(os.path.join(target_path, product_basename))

        except EnvironmentError as _error:
            raise Error("unable to retrieve product '%s' (%s) [%s]" % (product.core.product_name, product.core.uuid,
                                                                       _error))

        return paths

    def size(self, product_path):
        return util.product_size(os.path.join(self._root, product_path))

    def delete(self, product_path, properties):
        product_path = os.path.join(self._root, product_path)
        if not os.path.lexists(product_path):
            # If the product does not exist, do not consider this an error.
            return

        try:
            tmp_root = self.get_tmp_root(properties)
            with util.TemporaryDirectory(prefix=".remove-", suffix="-%s" % properties.core.uuid.hex,
                                         dir=tmp_root) as tmp_path:

                # Move product into the temporary directory. When the temporary directory will be removed at the end of
                # this scope, the product will be removed along with it.
                assert properties.core.physical_name == os.path.basename(product_path)
                os.rename(product_path, os.path.join(tmp_path, os.path.basename(product_path)))

        except EnvironmentError as _error:
            raise Error("unable to remove product '%s' (%s) [%s]" % (properties.core.product_name, properties.core.uuid,
                                                                     _error))

    def move(self, product, archive_path, paths=None):
        # Ignore if product already there
        if product.core.archive_path == archive_path:
            return paths

        # Make target archive path
        abs_archive_path = os.path.realpath(os.path.join(self._root, archive_path))
        util.make_path(abs_archive_path)

        # Move files there
        product_path = os.path.join(self._root, product.core.archive_path, product.core.physical_name)
        os.rename(product_path, os.path.join(abs_archive_path, product.core.physical_name))

        # Optionally rewrite (local) paths
        if paths is not None:
            paths = [os.path.join(self._root, archive_path,
                                  os.path.relpath(path, os.path.join(self._root, product.core.archive_path)))
                     for path in paths]
        return paths



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import logging
import multiprocessing
import sys

try:
    from tqdm import tqdm as bar
except ImportError:
    def bar(range, total=None, disable=None):
        return range

import muninn
from muninn.struct import Struct

from muninn.tools.utils import Processor, create_parser, parse_args_and_run

logger = logging.getLogger(__name__)

ACTIONS = [
    'ingest',
    'pull',
    'retype',
]


class UpdateProcessor(Processor):

    def __init__(self, args):
        super(UpdateProcessor, self).__init__(args)
        self.action = args.action
        self.argument = args.argument
        self.disable_hooks = args.disable_hooks
        self.use_current_path = args.keep
        self.verify_hash = args.verify_hash

    def perform_operation(self, archive, product):
        if self.action == 'ingest':
            logger.debug('running update:ingest on %s ' % product.core.product_name)
            archive.rebuild_properties(product.core.uuid, disable_hooks=self.disable_hooks,
                                       use_current_path=self.use_current_path)

        elif self.action == 'pull':
            logger.debug('running update:pull on %s ' % product.core.product_name)
            archive.rebuild_pull_properties(product.core.uuid, verify_hash=self.verify_hash,
                                            disable_hooks=self.disable_hooks,
                                            use_current_path=self.use_current_path)

        elif self.action == 'retype':
            if self.argument is not None:
                archive.update_properties(Struct({'core': {'product_type': self.argument}}), product.core.uuid)
            else:
                print('missing argument for retype action')
                sys.exit(1)


def update(args):
    filter = []
    if args.action in ['ingest', 'pull']:
        # only get products that are locally archived
        filter.append("(is_defined(archive_path) or is_defined(remote_url))")
    if args.action == 'pull':
        # only get products with a remote_url
        filter.append("is_defined(remote_url)")
    if args.expression:
        if len(filter) == 0:
            filter.append(args.expression)
        else:
            filter.append("(%s)" % args.expression)
    expression = " and ".join(filter)

    processor = UpdateProcessor(args)
    with muninn.open(args.archive) as archive:
        if args.action in ['ingest', 'pull']:
            # we only need the uuid and the product_name
            products = archive.search(expression, property_names=['uuid', 'product_name'])
        else:
            products = archive.search(expression, namespaces=archive.namespaces())
        if args.parallel:
            if args.processes is not None:
                pool = multiprocessing.Pool(args.processes)
            else:
                pool = multiprocessing.Pool()
            list(bar(pool.imap(processor, products), total=len(products), disable=None))
            pool.close()
            pool.join()
        else:
            for product in bar(products, disable=None):
                processor.perform_operation(archive, product)

    return 0


def main():
    parser = create_parser(description="""Updates properties of existing products.
        This is an archive maintenance tool, meant to be used when the archive structure has changed.
        Use with care!""", parallel=True)
    parser.add_argument("--disable-hooks", action="store_true",
                        help="do not run the hooks associated with the action")
    parser.add_argument("--verify-hash", action="store_true",
                        help="verify the hash of the product after a `pull` update")
    parser.add_argument("-k", "--keep", action="store_true",
                        help="do not attempt to relocate the product to the location specified in the "
                             "product type plug-in (useful for read-only archives)")
    parser.add_argument("action", metavar="ACTION", choices=ACTIONS, help="action name (%s)" % ', '.join(ACTIONS))
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    parser.add_argument("expression", metavar="EXPRESSION", default="", help="expression to select products")
    parser.add_argument("argument", metavar="ARGUMENT", nargs='?', help="product type for retype action")
    return parse_args_and_run(parser, update)


if __name__ == '__main__':
    main()
